"""Classes and functions for event subscription management."""

import ssl
import threading
from pathlib import Path
from typing import Callable, Optional

import paho.mqtt.client as mqtt
import stomp
import urllib3.util
import yaml
from waldur_api_client.api.event_subscriptions import (
    event_subscriptions_destroy,
)
from waldur_api_client.errors import UnexpectedStatus
from waldur_api_client.models.event_subscription import EventSubscription

from waldur_site_agent.backend import logger
from waldur_site_agent.common import utils
from waldur_site_agent.common.structures import Offering
from waldur_site_agent.event_processing import handlers
from waldur_site_agent.event_processing.listener import WaldurListener, connect_to_stomp_server
from waldur_site_agent.event_processing.structures import UserData

WALDUR_LISTENER_NAME = "waldur-listener"
OBJECT_TYPE_TO_HANDLER_STOMP = {
    "order": handlers.on_order_message_stomp,
    "user_role": handlers.on_user_role_message_stomp,
    "resource": handlers.on_resource_message_stomp,
    "importable_resources": handlers.on_importable_resources_message_stomp,
    "service_account": handlers.on_account_message_stomp,
    "course_account": handlers.on_account_message_stomp,
}
PID_FILE_PATH = "/var/run/waldur_site_agent.pid"


class EventSubscriptionManager:
    """Class responsible for management over event subscriptions and MQTT clients."""

    def __init__(
        self,
        offering: Offering,
        on_connect_callback: Optional[Callable] = None,
        on_message_callback: Optional[Callable] = None,
        user_agent: str = "",
        observable_object_type: str = "",
        global_proxy: str = "",
    ) -> None:
        """Constructor."""
        self.waldur_rest_client = utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl, global_proxy
        )
        self.offering = offering
        self.user_agent = user_agent
        self.on_connect_callback = on_connect_callback
        self.on_message_callback = on_message_callback
        self.observable_object_type = observable_object_type

    def _read_pid_file(self) -> dict:
        content = {}
        logger.info("Reading event subscriptions info from %s", PID_FILE_PATH)
        try:
            with Path(PID_FILE_PATH).open("r", encoding="utf-8") as pid_file:
                content = yaml.safe_load(pid_file)
        except FileNotFoundError:
            return content
        return content or {}

    def _write_event_subscription_info_to_pidfile(
        self, event_subscription: EventSubscription
    ) -> None:
        """Write event subscription to file."""
        pid_file_content = self._read_pid_file()
        logger.info(
            "Writing %s event subscription info to %s", self.observable_object_type, PID_FILE_PATH
        )

        with Path(PID_FILE_PATH).open("w+", encoding="utf-8") as pid_file:
            payload = {self.observable_object_type: event_subscription["uuid"]}
            pid_file_content.update(payload)
            yaml.dump(pid_file_content, pid_file)

    def _delete_event_subscription_from_pidfile(self) -> None:
        """Delete event subscription from pidfile."""
        pid_file_content = self._read_pid_file()
        logger.info(
            "Deleting %s event subscription info from %s if exists",
            self.observable_object_type,
            PID_FILE_PATH,
        )
        pid_file_content.pop(self.observable_object_type, None)
        with Path(PID_FILE_PATH).open("w+", encoding="utf-8") as pid_file:
            yaml.dump(pid_file_content, pid_file)

    def _setup_mqtt_consumer(self, event_subscription: EventSubscription) -> mqtt.Client:
        logger.info(
            "Setting up MQTT consumer for event subscription %s",
            event_subscription["uuid"],
        )
        mqtt_client = mqtt.Client(
            client_id=f"paho-mqtt-client-{event_subscription['uuid']}-{self.observable_object_type}",
            transport="websockets",
            callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
        )
        mqtt_client.ws_set_options(path="/rmqws", headers=None)
        vhost_name = event_subscription["user_uuid"]
        username = event_subscription["uuid"]
        username_full = f"{vhost_name}:{username}"
        password = self.offering.api_token
        mqtt_client.username_pw_set(username_full, password)
        if self.offering.websocket_use_tls:
            mqtt_client.tls_set()
        else:
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
            mqtt_client.tls_set_context(ssl_context)
            mqtt_client.tls_insecure_set(True)

        mqtt_client.on_connect = self.on_connect_callback
        mqtt_client.on_message = self.on_message_callback
        topic_postfix = self.observable_object_type
        userdata: UserData = {
            "event_subscription": event_subscription,
            "offering": self.offering,
            "user_agent": self.user_agent,
            "topic_postfix": topic_postfix,
        }
        mqtt_client.user_data_set(userdata)
        return mqtt_client

    def setup_stomp_connection(
        self,
        event_subscription: EventSubscription,
        custom_stomp_ws_host: Optional[str] = None,
        custom_stomp_ws_port: Optional[int] = None,
        custom_stomp_ws_path: Optional[str] = None,
    ) -> stomp.WSStompConnection:
        """Create a STOMP connection with the given parameters.

        Args:
            event_subscription (EventSubscription): Event subscription.
            custom_stomp_ws_host (Optional[str], optional): Custom host of the STOMP server.
                Defaults to None.
            custom_stomp_ws_port (Optional[int], optional): Custom port of the STOMP server.
                Defaults to None.
            custom_stomp_ws_path (Optional[str], optional): Custom path of the STOMP server.
                Defaults to None.

        Returns:
            stomp.WSStompConnection: The constructed connection
        """
        logger.info(
            "Setting up STOMP connection for event subscription %s",
            event_subscription.uuid.hex,
        )
        # Mapped to a vhost in RabbitMQ bound to a Waldur User object
        vhost_name = event_subscription.user_uuid.hex
        event_subscription_uuid = event_subscription.uuid.hex
        # Mapped to a username in RabbitMQ bound to the Waldur EventSubscription object
        username = event_subscription_uuid
        queue_name = (
            f"subscription_{event_subscription_uuid}_"
            f"offering_{self.offering.uuid}_{self.observable_object_type}"
        )

        stomp_host = custom_stomp_ws_host or urllib3.util.parse_url(self.offering.api_url).host
        stomp_port = (
            custom_stomp_ws_port or 443 if self.waldur_rest_client._verify_ssl else 80
        )  # TODO: Temporary workaround, improve later
        ws_path = custom_stomp_ws_path or "/rmqws-stomp"

        password = self.offering.api_token
        connection = stomp.WSStompConnection(
            host_and_ports=[(stomp_host, stomp_port)], ws_path=ws_path, vhost=vhost_name
        )
        if self.offering.websocket_use_tls:
            connection.set_ssl(for_hosts=[(stomp_host, stomp_port)])

        callback_function = OBJECT_TYPE_TO_HANDLER_STOMP[self.observable_object_type]
        connection.set_listener(
            WALDUR_LISTENER_NAME,
            WaldurListener(
                connection,
                queue_name,
                username,
                password,
                callback_function,
                self.offering,
                self.user_agent,
            ),
        )

        def create_stomp_thread(callback: Callable) -> threading.Thread:
            thread = threading.Thread(
                target=callback,
                group=None,
                name=f"waldur-{self.observable_object_type}-listener",
            )
            thread.daemon = True  # Don't let thread prevent termination
            thread.start()
            return thread

        connection.transport.override_threading(create_stomp_thread)
        return connection

    def start_stomp_connection(
        self,
        event_subscription: EventSubscription,
        connection: stomp.WSStompConnection,
    ) -> bool:
        """Start STOMP connection."""
        try:
            logger.info(
                "Starting STOMP connection for event subscription %s",
                event_subscription.uuid.hex,
            )
            connect_to_stomp_server(
                connection, event_subscription.uuid.hex, self.offering.api_token
            )
        except Exception as e:
            logger.error("Failed to start STOMP connection: %s", e)
            return False
        else:
            return True

    def stop_stomp_connection(self, connection: stomp.WSStompConnection) -> None:
        """Stop the STOMP connection."""
        connection.remove_listener(WALDUR_LISTENER_NAME)
        connection.disconnect()

    def start_mqtt_consumer(self, event_subscription: EventSubscription) -> Optional[mqtt.Client]:
        """Start MQTT consumer."""
        try:
            mqtt_host = urllib3.util.parse_url(self.offering.api_url).host
            mqtt_port = 443
            logger.info(
                "Starting consumer for %s (%s), MQTT WS address: ws://%s:%s",
                self.offering.name,
                self.offering.uuid,
                mqtt_host,
                mqtt_port,
            )

            mqtt_client = self._setup_mqtt_consumer(event_subscription)

            try:
                topic_name = (
                    f"subscription/{event_subscription['uuid']}/offering/"
                    f"{self.offering.uuid}/{self.observable_object_type}"
                )
                logger.info("Connecting the consumer to the mqtt server, topic: %s", topic_name)
                mqtt_client.connect(mqtt_host, mqtt_port)
            except (ConnectionRefusedError, TimeoutError) as e:
                logger.error("Failed to connect to MQTT broker: %s", e)
                return None
            except Exception as e:
                logger.error("Unexpected error while connecting to MQTT broker: %s", e)
                return None

            try:
                logger.info("Starting the consumer")
                mqtt_client.loop_start()
                logger.info("MQTT consumer started")
            except Exception as e:
                logger.error("Failed to start MQTT consumer loop: %s", e)
                mqtt_client.disconnect()
                return None
            else:
                return mqtt_client

        except Exception as e:
            logger.error("Failed to start MQTT consumer: %s", e)
            return None

    def stop_mqtt_consumer(self, mqttc: mqtt.Client) -> None:
        """Stop the MQTT consumer."""
        mqttc.loop_stop()
        mqttc.disconnect()

    def delete_event_subscription(self, event_subscription: EventSubscription) -> None:
        """Delete the event subscription."""
        try:
            logger.info("Deleting event subscription %s", event_subscription["uuid"])
            event_subscriptions_destroy.sync_detailed(
                uuid=event_subscription["uuid"], client=self.waldur_rest_client
            )
            logger.info("Event subscription deleted: %s", event_subscription["uuid"])
            self._delete_event_subscription_from_pidfile()
        except UnexpectedStatus as e:
            logger.error(
                "Failed to delete event subscription %s: %s", event_subscription["uuid"], e
            )



"""Functions shared between different backends."""

import calendar
import datetime
from zoneinfo import ZoneInfo

import yaml


def month_start(date: datetime.datetime) -> datetime.datetime:
    """Returns first day of a month for the date."""
    return datetime.datetime(day=1, month=date.month, year=date.year)


def month_end(date: datetime.datetime) -> datetime.date:
    """Returns last day of a month for the date."""
    days_in_month = calendar.monthrange(date.year, date.month)[1]
    return datetime.date(month=date.month, year=date.year, day=days_in_month)


def get_current_time_in_timezone(timezone_str: str = "") -> datetime.datetime:
    """Returns current time in the specified timezone.

    Or system timezone if no timezone is specified.
    """
    if timezone_str:
        try:
            tz = ZoneInfo(timezone_str)
            return datetime.datetime.now(tz)
        except Exception:
            return datetime.datetime.now()
    else:
        return datetime.datetime.now()


def format_current_month(timezone_str: str = "") -> tuple[str, str]:
    """Returns strings for start and end date of the current month."""
    today = get_current_time_in_timezone(timezone_str)
    start = month_start(today).strftime("%Y-%m-%dT00:00:00")
    end = month_end(today).strftime("%Y-%m-%dT23:59:59")
    return start, end


def get_usage_based_limits(resource_limits: dict) -> dict[str, int]:
    """Returns dictionary of limits for usage-based computing resources.

    The limits converted to SLURM-readable values.
    I.e. CPU-minutes, MB-minutes.
    """
    return {
        tres: data["limit"] * data.get("unit_factor", 1)
        for tres, data in resource_limits.items()
        if data["accounting_type"] == "usage" and data.get("limit") is not None
    }


def prettify_limits(limits: dict[str, int], slurm_tres: dict) -> str:
    """Makes limits human-readable."""
    limits_info = {
        slurm_tres[key]["label"]: f"{value} {slurm_tres[key]['measured_unit']}"
        for key, value in limits.items()
    }
    return yaml.dump(limits_info)


def sum_dicts(dict_list: list[dict]) -> dict:
    """Sums dictionaries by keys."""
    result_dict: dict[str, int] = {}

    # Iterate through each dictionary in the list
    for curr_dict in dict_list:
        # Sum values for each key in the current dictionary
        for key, value in curr_dict.items():
            result_dict[key] = result_dict.get(key, 0) + value

    return result_dict



"""Message listener module for Waldur STOMP plugin."""

import json
import time
from typing import Callable

import stomp.utils
from stomp.exception import StompException

from waldur_site_agent.backend import logger
from waldur_site_agent.common import structures


def connect_to_stomp_server(
    connection: stomp.StompConnection12, username: str, password: str
) -> None:
    """Connects the existing connection to the STOMP server."""
    while not connection.is_connected():
        try:
            logger.debug("Connecting to STOMP server as user %s", username)
            connection.connect(
                username,
                password,
                wait=True,
                headers={
                    "accept-version": "1.2",
                    "heart-beat": "10000,10000",  # Heartbeat configuration (client, server)
                },
            )
        except StompException as e:
            logger.error(
                "Failed to connect to the STOMP server, retrying in 10 seconds, reason: %s",
                e.__class__.__name__,
            )
            time.sleep(10)


class WaldurListener(stomp.ConnectionListener):
    """Message listener class for the STOMP plugin."""

    def __init__(
        self,
        conn: stomp.WSStompConnection,
        queue: str,
        username: str,
        password: str,
        on_message_callback: Callable,
        offering: structures.Offering,
        user_agent: str,
    ) -> None:
        """Constructor method."""
        self.queue = queue
        self.username = username
        self.password = password
        self.conn = conn
        self.on_message_callback = on_message_callback
        self.offering = offering
        self.user_agent = user_agent

    def on_error(self, frame: stomp.utils.Frame) -> None:
        """Error handler method."""
        logger.error("Received an error %s", frame.body)

    def on_message(self, frame: stomp.utils.Frame) -> None:
        """Message handler method."""
        logger.info("Received a message %s on queue %s", json.loads(frame.body), self.queue)
        try:
            self.on_message_callback(frame, self.offering, self.user_agent)
        except Exception as e:
            logger.exception(
                "Error processing message %s on queue %s: %s", frame.body, self.queue, e
            )

    def on_connected(self, _: stomp.utils.Frame) -> None:
        """Connection handler method."""
        logger.debug("Subscribing to %s", self.queue)
        self.conn.subscribe(
            destination=self.queue, id="waldur-subscription-", ack="auto"
        )  # TODO: try ack='client'

    def on_disconnected(self) -> None:
        """Disconnection handler method."""
        logger.debug("Disconnected from queue %s", self.queue)
        # Reconnecting after missing heartbeat
        connect_to_stomp_server(self.conn, self.username, self.password)



"""Common structures and data classes for the Waldur Site Agent.

This module defines the core data structures used throughout the agent:
- Configuration data classes for offerings and agent settings
- Enumerations for agent operational modes
- Structured representations of Waldur integration configuration

These structures provide type safety and clear interfaces for configuration
management across different agent components and backend plugins.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional


@dataclass
class Offering:
    """Configuration structure for a Waldur marketplace offering.

    This data class represents a single offering configuration from the agent's
    YAML configuration file. Each offering defines how the agent connects to
    Waldur and which backends to use for different operations.

    Attributes:
        name: Human-readable name for the offering
        api_url: Base URL for the Waldur API endpoint
        api_token: Authentication token for Waldur API access
        uuid: Unique identifier of the offering in Waldur
        backend_type: Legacy backend type identifier (deprecated)
        backend_settings: Backend-specific configuration parameters
        backend_components: Component definitions for the offering
        mqtt_enabled: Whether MQTT event processing is enabled
        websocket_use_tls: Whether to use TLS for websocket connections
        stomp_enabled: Whether STOMP event processing is enabled
        order_processing_backend: Backend name for order processing operations
        membership_sync_backend: Backend name for membership synchronization
        reporting_backend: Backend name for usage reporting
        username_management_backend: Backend name for username management
    """

    name: str = ""
    api_url: str = ""
    api_token: str = ""
    uuid: str = ""
    backend_type: str = ""
    backend_settings: dict = field(default_factory=dict)
    backend_components: dict = field(default_factory=dict)
    mqtt_enabled: bool = False
    websocket_use_tls: bool = True
    stomp_ws_host: Optional[str] = None
    stomp_ws_port: Optional[int] = None
    stomp_ws_path: Optional[str] = None
    stomp_enabled: bool = False
    order_processing_backend: Optional[str] = ""
    membership_sync_backend: Optional[str] = ""
    reporting_backend: Optional[str] = ""
    username_management_backend: str = "base"
    resource_import_enabled: bool = False
    verify_ssl: bool = True


class AgentMode(Enum):
    """Enumeration of operational modes for the Waldur Site Agent.

    The agent can operate in different modes, each handling specific aspects
    of the integration between Waldur and backend systems:

    - ORDER_PROCESS: Fetches orders from Waldur and creates/updates backend resources
    - REPORT: Collects usage data from backends and reports to Waldur
    - MEMBERSHIP_SYNC: Synchronizes user memberships between Waldur and backends
    - EVENT_PROCESS: Handles event-based processing via MQTT/STOMP
    """

    ORDER_PROCESS = "order_process"
    REPORT = "report"
    MEMBERSHIP_SYNC = "membership_sync"
    EVENT_PROCESS = "event_process"


@dataclass
class WaldurAgentConfiguration:
    """Complete configuration structure for the Waldur Site Agent.

    This data class holds the parsed and processed configuration from both
    command-line arguments and configuration files. It serves as the central
    configuration object used throughout the agent's operation.

    Attributes:
        waldur_offerings: List of offering configurations to process
        waldur_site_agent_mode: Current operational mode of the agent
        waldur_user_agent: HTTP User-Agent string for API requests
        waldur_site_agent_version: Version of the agent software
        sentry_dsn: Sentry DSN for error reporting (optional)
        timezone: Timezone for billing period calculations
        global_proxy: Global proxy URL for all API connections (optional)
    """

    waldur_offerings: list = field(default_factory=list)
    waldur_site_agent_mode: str = ""
    waldur_user_agent: str = ""
    waldur_site_agent_version: str = ""
    sentry_dsn: str = ""
    timezone: str = ""
    config_file_path: str = ""
    global_proxy: str = ""


class AccountType(Enum):
    """Enumeration of service account types for the Waldur Site Agent."""

    SERVICE_ACCOUNT = "service_account"
    COURSE_ACCOUNT = "course_account"



"""This module defines data structures used in event processing for the Waldur Site Agent."""

from __future__ import annotations

from typing import TypedDict

import paho.mqtt.client as mqtt
import stomp
from waldur_api_client.models.event_subscription import EventSubscription

from waldur_site_agent.common import structures as common_structures


class ObservableObject(TypedDict):
    """Represents an object that can be observed.

    Attributes:
        object_type (str): The model name of the observable object.
        object_uuid (str): The UUID of the object. If None, indicates that the subscriber
                          observes all existing objects of object_type.
    """

    object_type: str
    object_uuid: str


# A tuple containing MQTT client, subscription, and offering information.
MqttConsumer = tuple[mqtt.Client, EventSubscription, common_structures.Offering]

# A tuple of offering name and UUID used as a key for consumer mapping.
MqttConsumerKey = tuple[str, str]

# A tuple of offering name and UUID used as a key for connection mapping.
StompConsumerKey = tuple[str, str]

# A tuple containing STOMP connection, subscription, and offering information.
StompConsumer = tuple[stomp.WSStompConnection, EventSubscription, common_structures.Offering]

# A dictionary mapping consumer keys to lists of MQTT consumers.
MqttConsumersMap = dict[MqttConsumerKey, list[MqttConsumer]]
StompConsumersMap = dict[StompConsumerKey, list[StompConsumer]]


class UserData(TypedDict):
    """Represents user data for event handling and processing."""

    event_subscription: EventSubscription
    offering: common_structures.Offering
    user_agent: str
    topic_postfix: str


class UserRoleMessage(TypedDict):
    """Represents a message about user role changes in a project.

    Attributes:
        user_uuid (str, optional): The UUID of the user whose role is being modified.
        user_username (str, optional): The username of the user whose role is being modified.
        project_uuid (str): The UUID of the project where the role change occurred.
        project_name (str): The name of the project where the role change occurred.
        role_name (str): The name of the role that was granted or revoked.
        granted (bool, optional): True if the role was granted, False if it was revoked.
    """

    user_uuid: str | None
    user_username: str | None
    project_uuid: str
    project_name: str
    role_name: str
    granted: bool | None


class ResourceMessage(TypedDict):
    """Represents a message for a resource processing."""

    resource_uuid: str


class OrderMessage(TypedDict):
    """Represents a message for an order processing."""

    order_uuid: str
    order_state: str


class BackendResourceRequestMessage(TypedDict):
    """Represents a message for a backend resource request."""

    backend_resource_request_uuid: str


class AccountMessage(TypedDict):
    """Represents a message for service account processing."""

    account_uuid: str
    account_username: str
    scope_type: str
    project_uuid: str
    project_name: str
    action: str



"""Base module for different supported backends."""

import logging
import sys
from enum import Enum

ACCOUNT_NAME_REGEX = "a-zA-Z0-9-_"


class BackendType(Enum):
    """Enum for backend types."""

    SLURM = "slurm"
    MOAB = "moab"
    MUP = "mup"
    CUSTOM = "custom"
    UNKNOWN = "unknown"


console_handler = logging.StreamHandler(sys.stdout)
logger = logging.getLogger(__name__)

formatter = logging.Formatter("[%(levelname)s] [%(asctime)s] [%(threadName)s] %(message)s")
console_handler.setFormatter(formatter)

blank_handler = logging.StreamHandler(sys.stdout)
blank_handler.setLevel(logging.INFO)
blank_handler.setFormatter(logging.Formatter(fmt=""))

logger.addHandler(console_handler)
logger.setLevel(logging.INFO)



"""Shared utility functions for Waldur Site Agent modules.

This module provides common functionality used across different agent components:
- Configuration loading and parsing from CLI arguments and YAML files
- Waldur API client creation and authentication
- Backend discovery and initialization via entry points
- Component loading and synchronization with Waldur offerings
- Diagnostic and utility functions for system health checks
- Error handling utilities for resource management

The module handles the initialization of the agent's core configuration
and provides the plugin discovery mechanism that allows backends to be
automatically detected and loaded via Python entry points.
"""

import argparse
import sys
from pathlib import Path
from typing import Optional
from uuid import UUID

import yaml
from httpx import TimeoutException
from waldur_api_client import AuthenticatedClient
from waldur_api_client.api.marketplace_offering_users import (
    marketplace_offering_users_begin_creating,
    marketplace_offering_users_list,
    marketplace_offering_users_partial_update,
    marketplace_offering_users_set_ok,
    marketplace_offering_users_set_pending_account_linking,
    marketplace_offering_users_set_pending_additional_validation,
)
from waldur_api_client.api.marketplace_orders import marketplace_orders_list
from waldur_api_client.api.marketplace_provider_offerings import (
    marketplace_provider_offerings_create_offering_component,
    marketplace_provider_offerings_retrieve,
    marketplace_provider_offerings_update_offering_component,
)
from waldur_api_client.api.marketplace_provider_resources import (
    marketplace_provider_resources_set_as_erred,
    marketplace_provider_resources_set_limits,
)
from waldur_api_client.api.marketplace_resources import marketplace_resources_list
from waldur_api_client.api.users import users_me_retrieve
from waldur_api_client.errors import UnexpectedStatus
from waldur_api_client.models import (
    MarketplaceResourcesListStateItem,
    ResourceSetLimitsRequest,
    ResourceSetStateErredRequest,
)
from waldur_api_client.models.billing_type_enum import BillingTypeEnum
from waldur_api_client.models.marketplace_orders_list_state_item import (
    MarketplaceOrdersListStateItem,
)
from waldur_api_client.models.offering_component import OfferingComponent
from waldur_api_client.models.offering_component_request import OfferingComponentRequest
from waldur_api_client.models.offering_user import OfferingUser
from waldur_api_client.models.offering_user_state import OfferingUserState
from waldur_api_client.models.offering_user_state_transition_request import (
    OfferingUserStateTransitionRequest,
)
from waldur_api_client.models.patched_offering_user_request import PatchedOfferingUserRequest
from waldur_api_client.models.resource import Resource as WaldurResource
from waldur_api_client.models.update_offering_component_request import (
    UpdateOfferingComponentRequest,
)
from waldur_api_client.models.user import User
from waldur_api_client.models.username_generation_policy_enum import UsernameGenerationPolicyEnum

from waldur_site_agent.backend import (
    BackendType,
    logger,
)
from waldur_site_agent.backend import exceptions as backend_exceptions
from waldur_site_agent.backend.backends import (
    AbstractUsernameManagementBackend,
    BaseBackend,
    UnknownBackend,
    UnknownUsernameManagementBackend,
)
from waldur_site_agent.backend.exceptions import BackendError
from waldur_site_agent.common import WALDUR_SITE_AGENT_VERSION, pagination, structures

# Handle different Python versions
if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points, version
else:
    from importlib_metadata import entry_points, version


BACKENDS: dict[str, tuple[type[BaseBackend], str, str]] = {
    entry_point.name: (
        entry_point.load(),
        entry_point.dist.name if entry_point.dist else entry_point.name,
        version(entry_point.dist.name) if entry_point.dist else "unknown",
    )
    for entry_point in entry_points(group="waldur_site_agent.backends")
}

USERNAME_BACKENDS: dict[str, tuple[type[AbstractUsernameManagementBackend], str, str]] = {
    entry_point.name: (
        entry_point.load(),
        entry_point.dist.name if entry_point.dist else entry_point.name,
        version(entry_point.dist.name) if entry_point.dist else "unknown",
    )
    for entry_point in entry_points(group="waldur_site_agent.username_management_backends")
}

backend_packages = [
    {"package": package_distro, "entry_point": package_name, "version": package_version}
    for package_name, (_, package_distro, package_version) in BACKENDS.items()
]
username_backend_packages = [
    {"package": package_distro, "entry_point": package_name, "version": package_version}
    for package_name, (_, package_distro, package_version) in USERNAME_BACKENDS.items()
]
waldur_api_client_info = [
    {
        "package": "waldur-api-client",
        "entry_point": "waldur-api-client",
        "version": version("waldur-api-client"),
    }
]

DEPENDENCIES = backend_packages + username_backend_packages + waldur_api_client_info


def get_client(
    api_url: str,
    access_token: str,
    agent_header: Optional[str] = None,
    verify_ssl: bool = True,
    proxy: Optional[str] = None,
) -> AuthenticatedClient:
    """Create an authenticated Waldur API client.

    Args:
        api_url: Base URL for the Waldur API (e.g., 'https://waldur.example.com/api/')
        access_token: Authentication token for API access
        agent_header: Optional User-Agent string for HTTP requests
        verify_ssl: Whether or not to verify SSL certificates
        proxy: Optional proxy URL (e.g., 'socks5://localhost:12345')

    Returns:
        Configured AuthenticatedClient instance ready for API calls
    """
    headers = {"User-Agent": agent_header} if agent_header else {}
    url = api_url.rstrip("/api")

    # Configure httpx args with proxy if specified
    httpx_args = {}
    if proxy:
        httpx_args["proxy"] = proxy

    return AuthenticatedClient(
        base_url=url,
        token=access_token,
        timeout=600,
        headers=headers,
        verify_ssl=verify_ssl,
        httpx_args=httpx_args,
    )


def is_uuid(value: str) -> bool:
    """Validate if a string represents a valid UUID.

    Args:
        value: String to validate as UUID

    Returns:
        True if the string is a valid UUID, False otherwise
    """
    try:
        UUID(value)
        return True
    except ValueError:
        return False


def load_configuration(
    config_file_path: str, user_agent_suffix: str = "generic"
) -> structures.WaldurAgentConfiguration:
    """Load configuration from YAML file.

    Args:
        config_file_path: Path to the YAML configuration file
        user_agent_suffix: Suffix to add to the user agent string (e.g., "sync", "order-process")

    Returns:
        Configuration object with offerings loaded from file

    Raises:
        FileNotFoundError: If the configuration file cannot be found
        yaml.YAMLError: If the configuration file is malformed
    """
    configuration = structures.WaldurAgentConfiguration()

    with Path(config_file_path).open(encoding="UTF-8") as stream:
        config = yaml.safe_load(stream)
        offering_list = config["offerings"]
        waldur_offerings = [
            structures.Offering(
                name=offering_info["name"],
                api_url=offering_info["waldur_api_url"],
                api_token=offering_info["waldur_api_token"],
                uuid=offering_info["waldur_offering_uuid"],
                backend_type=offering_info["backend_type"].lower(),
                backend_settings=offering_info["backend_settings"],
                backend_components=offering_info["backend_components"],
                mqtt_enabled=offering_info.get("mqtt_enabled", False),
                stomp_enabled=offering_info.get("stomp_enabled", False),
                websocket_use_tls=offering_info.get("websocket_use_tls", True),
                stomp_ws_host=offering_info.get("stomp_ws_host"),
                stomp_ws_port=offering_info.get("stomp_ws_port"),
                stomp_ws_path=offering_info.get("stomp_ws_path"),
                username_management_backend=offering_info.get(
                    "username_management_backend", "base"
                ),
                order_processing_backend=offering_info.get("order_processing_backend", ""),
                membership_sync_backend=offering_info.get("membership_sync_backend", ""),
                reporting_backend=offering_info.get("reporting_backend", ""),
                resource_import_enabled=offering_info.get("resource_import_enabled", False),
                verify_ssl=offering_info.get("verify_ssl", True),
            )
            for offering_info in offering_list
        ]
        configuration.waldur_offerings = waldur_offerings

        # Handle Sentry configuration - initialize if DSN is provided
        sentry_dsn = config.get("sentry_dsn")
        if sentry_dsn:
            configuration.sentry_dsn = sentry_dsn
            import sentry_sdk  # noqa: PLC0415

            sentry_sdk.init(dsn=sentry_dsn)

        timezone = config.get("timezone", "UTC")
        configuration.timezone = timezone

        # Handle global proxy configuration
        global_proxy = config.get("global_proxy", "")
        configuration.global_proxy = global_proxy

    # Set version and user agent for all configurations
    configuration.waldur_site_agent_version = WALDUR_SITE_AGENT_VERSION
    configuration.waldur_user_agent = (
        f"waldur-site-agent-{user_agent_suffix}/{WALDUR_SITE_AGENT_VERSION}"
    )
    configuration.config_file_path = config_file_path

    return configuration


def init_configuration() -> structures.WaldurAgentConfiguration:
    """Initialize agent configuration from CLI arguments and config file.

    Parses command-line arguments, loads the YAML configuration file,
    and creates offering configurations. Also initializes Sentry if
    configured and sets up user agent strings for different modes.

    Returns:
        Complete agent configuration with all offerings and settings

    Raises:
        FileNotFoundError: If the configuration file cannot be found
        yaml.YAMLError: If the configuration file is malformed
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--mode",
        "-m",
        help="Agent mode, choices: order_process, report "
        "membership_sync and event_process; default is order_process",
        choices=["order_process", "report", "membership_sync", "event_process"],
        default="order_process",
    )

    parser.add_argument(
        "--config-file",
        "-c",
        help="Path to the config file with provider settings;"
        "default is waldur-site-agent-config.yaml",
        dest="config_file_path",
        default="waldur-site-agent-config.yaml",
        required=False,
    )

    cli_args = parser.parse_args()
    config_file_path = cli_args.config_file_path
    agent_mode = cli_args.mode

    logger.info("Using %s as a config source", config_file_path)

    # Load base configuration with mode-specific user agent
    configuration = load_configuration(config_file_path, user_agent_suffix=agent_mode)

    # Add CLI-specific configuration
    configuration.waldur_site_agent_mode = agent_mode

    return configuration


def get_backend_for_offering(
    offering: structures.Offering, backend_type_key: str
) -> tuple[BaseBackend, str]:
    """Create and initialize a backend instance for the specified offering.

    Uses the plugin discovery system to find and instantiate the appropriate
    backend class based on the offering's configuration.

    Args:
        offering: The offering configuration
        backend_type_key: Key to determine which backend type to use
                         (e.g., 'order_processing_backend', 'reporting_backend')

    Returns:
        Initialized backend instance with version as a second value,
        or UnknownBackend if type not supported
    """
    backend_type = getattr(offering, backend_type_key, "")
    backend_info = BACKENDS.get(backend_type)
    if not backend_info:
        logger.error("Unsupported backend type for %s: %s", backend_type_key, backend_type)
        return UnknownBackend(), "unknown"

    backend_class, dist_name, dist_version = backend_info

    logger.info(
        "Using %s backend (distro %s, version %s) for %s offering",
        backend_type,
        dist_name,
        dist_version,
        offering.name,
    )

    return backend_class(offering.backend_settings, offering.backend_components), dist_version


def mark_waldur_resources_as_erred(
    waldur_rest_client: AuthenticatedClient,
    resources: list[WaldurResource],
    error_details: dict[str, str],
) -> None:
    """Mark multiple resources as ERRED in Waldur with error details.

    This utility function handles batch error reporting for resources
    that have encountered processing failures.

    Args:
        waldur_rest_client: Authenticated Waldur API client
        resources: List of resources to mark as erred
        error_details: Dictionary containing 'error_message' and 'error_traceback'
    """
    logger.info("Marking Waldur resources as ERRED")
    for resource in resources:
        logger.info("Marking %s resource as ERRED", resource)
        try:
            request_body = ResourceSetStateErredRequest(
                error_message=error_details.get("error_message", ""),
                error_traceback=error_details.get("error_traceback", ""),
            )
            marketplace_provider_resources_set_as_erred.sync_detailed(
                uuid=resource.uuid.hex, client=waldur_rest_client, body=request_body
            )
        except UnexpectedStatus as e:
            logger.exception(
                "Waldur REST client error while setting resource state to Erred %s: %s",
                resource.backend_id,
                e,
            )


def load_offering_components() -> None:
    """Load and create offering components in Waldur from configuration.

    This function reads the agent configuration and creates or updates
    offering components in Waldur to match the backend component definitions.
    Used during initial setup or when component definitions change.
    """
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        logger.info("Processing %s offering", offering.name)
        waldur_rest_client = get_client(
            offering.api_url,
            offering.api_token,
            configuration.waldur_user_agent,
            offering.verify_ssl,
            configuration.global_proxy,
        )

        load_components_to_waldur(
            waldur_rest_client,
            offering.uuid,
            offering.name,
            offering.backend_components,
        )


def extend_backend_components(
    offering: structures.Offering, waldur_offering_components: list[OfferingComponent]
) -> None:
    """Synchronize local configuration with Waldur offering components.

    Fetches component definitions from Waldur and adds any missing components
    to the local offering configuration. This ensures consistency between
    the agent's configuration and Waldur's offering definition.

    Args:
        offering: Local offering configuration to extend
        waldur_offering_components: Component definitions from Waldur
    """
    logger.info("Loading Waldur components to the local config")
    remote_components: dict[str, OfferingComponent] = {
        item.type_: item for item in waldur_offering_components
    }
    missing_component_types = set(remote_components.keys()) - set(
        offering.backend_components.keys()
    )

    if missing_component_types:
        logger.info("Component types to add: %s", ", ".join(missing_component_types))
    else:
        logger.info("All remote components already exist in backend configuration, nothing to add")
    for missing_component_type in missing_component_types:
        logger.info("Loading %s", missing_component_type)
        remote_component_info = remote_components[missing_component_type]
        component_info = {
            "limit": remote_component_info.limit_amount,
            "measured_unit": remote_component_info.measured_unit,
            "unit_factor": remote_component_info.unit_factor or 1,
            "accounting_type": remote_component_info.billing_type,
            "label": remote_component_info.name,
        }
        offering.backend_components[missing_component_type] = component_info


def load_components_to_waldur(
    waldur_rest_client: AuthenticatedClient,
    offering_uuid: str,
    offering_name: str,
    components: dict,
) -> None:
    """Create or update offering components in Waldur.

    Processes the component definitions from the configuration and creates
    or updates the corresponding components in Waldur. Handles both new
    component creation and limit updates for existing components.

    Args:
        waldur_rest_client: Authenticated Waldur API client
        offering_uuid: UUID of the target offering in Waldur
        offering_name: Name of the offering (for logging)
        components: Dictionary of component definitions from configuration
    """
    logger.info(
        "Creating offering components data for the following resources: %s",
        ", ".join(components.keys()),
    )
    waldur_offering = marketplace_provider_offerings_retrieve.sync(
        client=waldur_rest_client, uuid=offering_uuid
    )
    waldur_offering_components: dict[str, OfferingComponent] = {
        component.type_: component for component in waldur_offering.components
    }
    for component_type, component_info in components.items():
        try:
            limit_amount = component_info.get("limit")
            accounting_type = component_info["accounting_type"]
            label = component_info["label"]

            if component_type in waldur_offering_components:
                if component_info["accounting_type"] == "usage":
                    existing_component = waldur_offering_components[component_type]
                    logger.info(
                        "Offering component %s already exists, updating limit from %s to %s %s.",
                        component_type,
                        existing_component.limit_amount,
                        component_info.get("limit"),
                        component_info["measured_unit"],
                    )
                    component_update_body = UpdateOfferingComponentRequest(
                        uuid=existing_component.uuid.hex,
                        billing_type=BillingTypeEnum(accounting_type),
                        type_=component_type,
                        name=label,
                        measured_unit=component_info["measured_unit"],
                        limit_amount=limit_amount,
                    )
                    marketplace_provider_offerings_update_offering_component.sync_detailed(
                        client=waldur_rest_client,
                        uuid=existing_component.uuid,
                        body=component_update_body,
                    )
                else:
                    logger.info(
                        "Offering component %s already exists, skipping creation.",
                        component_type,
                    )
            else:
                logger.info(
                    "Creating offering component %s with type %s and limit %s %s.",
                    component_type,
                    component_info["accounting_type"],
                    component_info.get("limit"),
                    component_info["measured_unit"],
                )
                component = OfferingComponentRequest(
                    billing_type=BillingTypeEnum(accounting_type),
                    type_=component_type,
                    name=label,
                    measured_unit=component_info["measured_unit"],
                    limit_amount=limit_amount,
                )
                marketplace_provider_offerings_create_offering_component.sync_detailed(
                    client=waldur_rest_client, uuid=offering_uuid, body=component
                )
        except Exception as e:
            logger.info(
                "Unable to create or update a component %s for offering %s (%s):",
                component_info["label"],
                offering_name,
                offering_uuid,
            )
            logger.exception(e)


def get_current_user_from_client(waldur_rest_client: AuthenticatedClient) -> User:
    """Retrieve current authenticated user information from Waldur.

    Args:
        waldur_rest_client: Authenticated Waldur API client

    Returns:
        User object containing current user details and permissions
    """
    return users_me_retrieve.sync(client=waldur_rest_client)


def diagnostics() -> bool:
    """Perform comprehensive system diagnostics for all offerings.

    Checks connectivity to Waldur, validates offering configurations,
    tests backend availability, and reports system status. This function
    is used by the diagnostic command to verify agent setup.

    Returns:
        True if all diagnostics pass, False if any issues are detected
    """
    configuration = init_configuration()
    logger.info("-" * 10 + "DIAGNOSTICS START" + "-" * 10)
    logger.info("Provided settings:")
    format_string = "{:<30} = {:<10}"

    if structures.AgentMode.ORDER_PROCESS.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - "
            "pulling orders from Waldur and creating resources in backend",
            structures.AgentMode.ORDER_PROCESS.name,
        )
    if structures.AgentMode.REPORT.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - pushing usage data to Waldur",
            structures.AgentMode.REPORT.name,
        )
    if structures.AgentMode.MEMBERSHIP_SYNC.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - pushing membership data to Waldur",
            structures.AgentMode.MEMBERSHIP_SYNC.name,
        )

    if structures.AgentMode.EVENT_PROCESS.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - processing data from Waldur in event-based approach",
            structures.AgentMode.EVENT_PROCESS.name,
        )

    for offering in configuration.waldur_offerings:
        format_string = "{:<30} = {:<10}"
        offering_uuid = offering.uuid
        offering_name = offering.name
        offering_api_url = offering.api_url
        offering_api_token = offering.api_token

        logger.info(format_string.format("Offering name", offering_name))
        logger.info(format_string.format("Offering UUID", offering_uuid))
        logger.info(format_string.format("Waldur API URL", offering_api_url))
        logger.info(format_string.format("SENTRY_DSN", str(configuration.sentry_dsn)))

        waldur_rest_client = get_client(
            offering_api_url,
            offering_api_token,
            configuration.waldur_user_agent,
            offering.verify_ssl,
        )

        try:
            current_user = get_current_user_from_client(waldur_rest_client)
            print_current_user(current_user)
            offering_data = marketplace_provider_offerings_retrieve.sync(
                client=waldur_rest_client, uuid=offering_uuid
            )
            logger.info("Offering uuid: %s", offering_data.uuid)
            logger.info("Offering name: %s", offering_data.name)
            logger.info("Offering org: %s", offering_data.customer_name)
            logger.info("Offering state: %s", offering_data.state)

            logger.info("Offering components:")
            format_string = "{:<10} {:<10} {:<10} {:<10}"
            headers = ["Type", "Name", "Unit", "Limit"]
            logger.info(format_string.format(*headers))
            components = [
                [
                    component.type_,
                    component.name,
                    component.measured_unit,
                    component.limit_amount or "",
                ]
                for component in offering_data.components
            ]
            for component in components:
                logger.info(format_string.format(*component))

            logger.info("")
        except UnexpectedStatus as err:
            logger.error("Unable to fetch offering data, reason: %s", err)

        logger.info("")
        try:
            orders = marketplace_orders_list.sync(
                client=waldur_rest_client,
                offering_uuid=offering_uuid,
                state=[
                    MarketplaceOrdersListStateItem.PENDING_PROVIDER,
                    MarketplaceOrdersListStateItem.EXECUTING,
                ],
                page_size=100,
            )
            logger.info("Active orders:")
            format_string = "{:<10} {:<10} {:<10}"
            headers = ["Project", "Type", "State"]
            logger.info(format_string.format(*headers))
            for order in orders:
                logger.info(format_string.format(order.project_name, order.type_, order.state))
        except UnexpectedStatus as err:
            logger.error("Unable to fetch orders, reason: %s", err)

        backend, _ = get_backend_for_offering(offering, "order_processing_backend")

        if not backend.diagnostics():
            return False

    logger.info("-" * 10 + "DIAGNOSTICS END" + "-" * 10)
    return True


def create_homedirs_for_offering_users() -> None:
    """Create home directories for all offering users.

    This utility function creates home directories for users associated
    with offerings that have home directory creation enabled. Currently
    supports SLURM backends with configurable umask settings.
    """
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        # Feature is exclusive for SLURM temporarily
        if offering.backend_type != BackendType.SLURM.value or not offering.backend_settings.get(
            "enable_user_homedir_account_creation", True
        ):
            continue

        logger.info("Creating homedirs for %s offering users", offering.name)

        waldur_rest_client = get_client(
            offering.api_url,
            offering.api_token,
            configuration.waldur_user_agent,
            offering.verify_ssl,
            configuration.global_proxy,
        )
        offering_users = pagination.get_all_paginated(
            marketplace_offering_users_list.sync,
            client=waldur_rest_client,
            offering_uuid=[offering.uuid],
            state=[OfferingUserState.OK],
            is_restricted=False,
        )

        offering_user_usernames: set[str] = {
            offering_user.username for offering_user in offering_users
        }
        umask = offering.backend_settings.get("homedir_umask", "0700")
        offering_backend, _ = get_backend_for_offering(offering, "order_processing_backend")
        offering_backend.create_user_homedirs(offering_user_usernames, umask)


def print_current_user(current_user: User) -> None:
    """Log detailed information about a Waldur user.

    Displays user details including username, full name, staff status,
    and all associated permissions with their scopes and expiration times.

    Args:
        current_user: User object to display information for
    """
    logger.info("Current user username: %s", current_user.username)
    logger.info("Current user full name: %s", current_user.full_name)
    logger.info("Current user is staff: %s", current_user.is_staff or False)
    if current_user.is_staff:
        return
    if current_user.permissions:
        logger.info("List of permissions:")
        for permission in current_user.permissions:
            logger.info("Role name: %s", permission.role_name)
            logger.info("Role description: %s", permission.role_description)
            logger.info("Scope type: %s", permission.scope_type)
            logger.info("Scope name: %s", permission.scope_name)
            logger.info("Scope UUID: %s", permission.scope_uuid)
            logger.info("Expiration time: %s", permission.expiration_time)
    else:
        logger.info("User has no role permissions.")


def get_username_management_backend(
    offering: structures.Offering,
) -> tuple[AbstractUsernameManagementBackend, str]:
    """Create username management backend instance for the offering.

    Uses the plugin discovery system to instantiate the appropriate
    username management backend based on the offering configuration.

    Args:
        offering: Offering configuration specifying the backend to use

    Returns:
        Username management backend instance, or UnknownUsernameManagementBackend
        if the specified backend is not available
    """
    username_management_setting = offering.username_management_backend

    if username_management_setting is None:
        logger.error(
            "No username_management_backend is set for offering %s, using the default one",
            offering.name,
        )
        return UnknownUsernameManagementBackend(), "unknown"

    backend_info = USERNAME_BACKENDS[username_management_setting]
    backend_class, dist_name, dist_version = backend_info

    logger.info(
        "Using %s backend (distro %s, version %s) for %s offering",
        username_management_setting,
        dist_name,
        dist_version,
        offering.name,
    )

    return backend_class(), dist_version


def update_offering_users(
    offering: structures.Offering,
    waldur_rest_client: AuthenticatedClient,
    offering_users: list[OfferingUser],
) -> bool:
    """Generate usernames for offering users and update their state accordingly.

    This method checks if the service provider is allowed to generate usernames
    and attempts to create usernames for users who don't have them assigned.

    Args:
        offering: The offering to process,
        waldur_rest_client: Authenticated Waldur API client
        offering_users: List of offering users to process

    Returns:
        True if any usernames were updated, False otherwise
    """
    if not offering_users:
        return False

    # Early return if username generation is not allowed
    if not _can_generate_usernames(offering, waldur_rest_client):
        return False

    username_management_backend, _ = get_username_management_backend(offering)

    # Skip processing if we have an unknown username management backend
    if isinstance(username_management_backend, UnknownUsernameManagementBackend):
        logger.debug("Skipping username processing - unknown username management backend")
        return False

    # Group users by their current state for efficient processing
    requested_users, pending_users = _group_users_by_state(offering_users)

    changed = False

    # Process requested users first
    if requested_users:
        changed |= _process_requested_users(
            requested_users, username_management_backend, waldur_rest_client, offering
        )

    # Then process pending users
    if pending_users:
        changed |= _process_pending_users(
            pending_users, username_management_backend, waldur_rest_client
        )

    return changed


def _can_generate_usernames(
    offering: structures.Offering, waldur_rest_client: AuthenticatedClient
) -> bool:
    """Check if usernames can be generated by the service provider."""
    offering_details = marketplace_provider_offerings_retrieve.sync(
        client=waldur_rest_client, uuid=offering.uuid
    )
    return (
        offering_details.plugin_options.username_generation_policy
        == UsernameGenerationPolicyEnum.SERVICE_PROVIDER
    )


def _group_users_by_state(
    offering_users: list[OfferingUser],
) -> tuple[list[OfferingUser], list[OfferingUser]]:
    """Group users by their processing state."""
    requested_users = []
    pending_users = []

    for user in offering_users:
        if user.state == OfferingUserState.REQUESTED:
            requested_users.append(user)
        elif user.state in {
            OfferingUserState.CREATING,
            OfferingUserState.PENDING_ACCOUNT_LINKING,
            OfferingUserState.PENDING_ADDITIONAL_VALIDATION,
        }:
            pending_users.append(user)

    return requested_users, pending_users


def _process_requested_users(
    users: list[OfferingUser],
    username_management_backend: AbstractUsernameManagementBackend,
    waldur_rest_client: AuthenticatedClient,
    offering: structures.Offering,
) -> bool:
    """Process users in REQUESTED state."""
    logger.info(
        "Generating usernames for %d requested offering users in offering %s",
        len(users),
        offering.uuid,
    )

    changed = False
    for user in users:
        try:
            logger.info(
                "Setting offering user %s (%s) state to CREATING",
                user.user_email,
                user.uuid,
            )
            marketplace_offering_users_begin_creating.sync_detailed(
                uuid=user.uuid, client=waldur_rest_client
            )

            if _update_user_username(user, username_management_backend, waldur_rest_client):
                changed = True

        except backend_exceptions.OfferingUserAccountLinkingRequiredError as e:
            _handle_account_linking_error(user, e, waldur_rest_client)
        except backend_exceptions.OfferingUserAdditionalValidationRequiredError as e:
            _handle_validation_error(user, e, waldur_rest_client)
        except Exception as e:
            logger.error(
                "Failed to generate username for offering user %s (%s): %s",
                user.user_email,
                user.username,
                e,
            )

    return changed


def _process_pending_users(
    users: list[OfferingUser],
    username_management_backend: AbstractUsernameManagementBackend,
    waldur_rest_client: AuthenticatedClient,
) -> bool:
    """Process users in pending states."""
    logger.info("Processing %d pending offering users", len(users))

    changed = False
    for user in users:
        try:
            logger.info(
                "Checking username for offering user %s (%s)",
                user.user_email,
                user.uuid,
            )

            if _update_user_username(user, username_management_backend, waldur_rest_client):
                changed = True

        except (
            backend_exceptions.OfferingUserAccountLinkingRequiredError,
            backend_exceptions.OfferingUserAdditionalValidationRequiredError,
        ):
            logger.info("Backend account is still in the pending state")
        except Exception as e:
            logger.error(
                "Failed to generate username for offering user %s (%s): %s",
                user.user_email,
                user.username,
                e,
            )

    return changed


def _update_user_username(
    offering_user: OfferingUser,
    username_management_backend: AbstractUsernameManagementBackend,
    waldur_rest_client: AuthenticatedClient,
) -> bool:
    """Update username for a single offering user."""
    username = username_management_backend.get_or_create_username(offering_user)
    if not username:
        return False

    logger.info(
        "Updating username for offering user %s (%s) to %s",
        offering_user.user_email,
        offering_user.uuid,
        username,
    )

    # Update local object and remote API
    offering_user.username = username
    payload = PatchedOfferingUserRequest(username=username)
    marketplace_offering_users_partial_update.sync(
        uuid=offering_user.uuid, client=waldur_rest_client, body=payload
    )

    logger.info("Setting offering user state to OK")
    marketplace_offering_users_set_ok.sync_detailed(
        uuid=offering_user.uuid, client=waldur_rest_client
    )

    return True


def _handle_account_linking_error(
    user: OfferingUser,
    error: backend_exceptions.OfferingUserAccountLinkingRequiredError,
    waldur_rest_client: AuthenticatedClient,
) -> None:
    """Handle account linking required error."""
    logger.warning(
        "Offering user %s (%s) requires user linking: %s",
        user.user_email,
        user.uuid,
        error,
    )

    comment_url = getattr(error, "comment_url", None) if hasattr(error, "comment_url") else None
    payload = OfferingUserStateTransitionRequest(comment=str(error), comment_url=comment_url)
    marketplace_offering_users_set_pending_account_linking.sync_detailed(
        uuid=user.uuid, client=waldur_rest_client, body=payload
    )


def _handle_validation_error(
    user: OfferingUser,
    error: backend_exceptions.OfferingUserAdditionalValidationRequiredError,
    waldur_rest_client: AuthenticatedClient,
) -> None:
    """Handle additional validation required error."""
    logger.warning(
        "Offering user %s (%s) requires additional validation: %s",
        user.user_email,
        user.uuid,
        error,
    )

    comment_url = getattr(error, "comment_url", None) if hasattr(error, "comment_url") else None
    payload = OfferingUserStateTransitionRequest(comment=str(error), comment_url=comment_url)
    marketplace_offering_users_set_pending_additional_validation.sync_detailed(
        uuid=user.uuid, client=waldur_rest_client, body=payload
    )


def sync_offering_users() -> None:
    """Process offering users for all configured offerings.

    This function retrieves offering users from Waldur and processes them
    according to the configured username management backend. It handles
    username generation and state transitions for offering users
    based on the backend's capabilities.
    """
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        logger.info("Processing offering users for %s", offering.name)

        waldur_rest_client = get_client(
            offering.api_url,
            offering.api_token,
            configuration.waldur_user_agent,
            offering.verify_ssl,
            configuration.global_proxy,
        )
        offering_users = pagination.get_all_paginated(
            marketplace_offering_users_list.sync,
            client=waldur_rest_client,
            offering_uuid=[offering.uuid],
            is_restricted=False,
        )
        update_offering_users(offering, waldur_rest_client, offering_users)


def sync_waldur_resource_limits(
    resource_backend: BaseBackend,
    waldur_rest_client: AuthenticatedClient,
    waldur_resource: WaldurResource,
) -> None:
    """Syncs resource limits between Waldur and the backend.

    The method is shared between utils and processors.
    """
    logger.info(
        "Syncing resource limits for resource %s (%s)",
        waldur_resource.name,
        waldur_resource.backend_id,
    )

    backend_limits = resource_backend.get_resource_limits(waldur_resource.backend_id)

    if not backend_limits:
        logger.warning("No limits found in the backend")
        return

    if waldur_resource.limits.additional_properties == backend_limits:
        logger.info("The limits are already in sync (%s), skipping", backend_limits)
        return

    # For now, we report all the limits
    logger.info("Changing resource limits from %s to %s", waldur_resource.limits, backend_limits)

    marketplace_provider_resources_set_limits.sync(
        uuid=waldur_resource.uuid.hex,
        client=waldur_rest_client,
        body=ResourceSetLimitsRequest(limits=backend_limits),
    )


def sync_resource_limits() -> None:
    """Report resource limits for the existing resources to Waldur."""
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        logger.info(
            "Processing resource limits for offering %s, backend plugin is %s",
            offering.name,
            offering.membership_sync_backend,
        )
        backend, _ = get_backend_for_offering(offering, "membership_sync_backend")
        logger.info("Using class %s as a backend", backend.__class__.__name__)
        waldur_rest_client = get_client(
            offering.api_url,
            offering.api_token,
            configuration.waldur_user_agent,
            offering.verify_ssl,
            configuration.global_proxy,
        )
        resources = pagination.get_all_paginated(
            marketplace_resources_list.sync,
            client=waldur_rest_client,
            offering_uuid=[offering.uuid],
            state=[MarketplaceResourcesListStateItem.OK, MarketplaceResourcesListStateItem.ERRED],
        )

        waldur_resources = [resource for resource in resources if resource.backend_id]
        logger.info("Processing limits for %s resource(s)", len(waldur_resources))
        for waldur_resource in waldur_resources:
            try:
                sync_waldur_resource_limits(backend, waldur_rest_client, waldur_resource)
            except (BackendError, UnexpectedStatus, TimeoutException) as e:
                logger.error(
                    "Failed to sync resource limits for %s, reason: %s", waldur_resource.name, e
                )



"""Abstract offering processors for different agent operational modes.

This module provides the core abstract base classes that define the interface
and common functionality for processing Waldur offerings. The processors handle
the integration between Waldur Mastermind and backend systems through different
operational patterns:

- OfferingOrderProcessor: Handles order lifecycle (create, update, terminate)
- OfferingMembershipProcessor: Manages user membership synchronization
- OfferingReportProcessor: Reports usage data from backends to Waldur

Each processor type implements specific aspects of the Waldur-backend integration
while sharing common patterns for error handling, retry logic, and API communication.
"""

from __future__ import annotations

import abc
import traceback
from time import sleep
from typing import Any, Optional

from waldur_api_client.api.backend_resource_requests import (
    backend_resource_requests_retrieve,
    backend_resource_requests_set_done,
    backend_resource_requests_set_erred,
    backend_resource_requests_start_processing,
)
from waldur_api_client.api.backend_resources import backend_resources_create, backend_resources_list
from waldur_api_client.api.component_user_usage_limits import component_user_usage_limits_list
from waldur_api_client.api.marketplace_component_usages import (
    marketplace_component_usages_list,
    marketplace_component_usages_set_usage,
    marketplace_component_usages_set_user_usage,
)
from waldur_api_client.api.marketplace_offering_users import (
    marketplace_offering_users_list,
)
from waldur_api_client.api.marketplace_orders import (
    marketplace_orders_approve_by_provider,
    marketplace_orders_list,
    marketplace_orders_retrieve,
    marketplace_orders_set_state_done,
    marketplace_orders_set_state_erred,
)
from waldur_api_client.api.marketplace_provider_offerings import (
    marketplace_provider_offerings_retrieve,
)
from waldur_api_client.api.marketplace_provider_resources import (
    marketplace_provider_resources_list,
    marketplace_provider_resources_refresh_last_sync,
    marketplace_provider_resources_retrieve,
    marketplace_provider_resources_set_as_ok,
    marketplace_provider_resources_set_backend_id,
    marketplace_provider_resources_set_backend_metadata,
    marketplace_provider_resources_set_limits,
    marketplace_provider_resources_team_list,
)
from waldur_api_client.api.marketplace_service_providers import (
    marketplace_service_providers_course_accounts_list,
    marketplace_service_providers_list,
    marketplace_service_providers_project_service_accounts_list,
)
from waldur_api_client.api.projects import projects_list
from waldur_api_client.errors import UnexpectedStatus
from waldur_api_client.models import (
    ComponentUsageCreateRequest,
    ComponentUsageItemRequest,
    CourseAccount,
    MarketplaceOfferingUsersListStateItem,
    ProjectServiceAccount,
    ResourceSetLimitsRequest,
    ServiceAccountState,
)
from waldur_api_client.models.agent_processor import AgentProcessor
from waldur_api_client.models.agent_service import AgentService
from waldur_api_client.models.backend_resource_request import BackendResourceRequest
from waldur_api_client.models.backend_resource_request_set_erred_request import (
    BackendResourceRequestSetErredRequest,
)
from waldur_api_client.models.component_usage import ComponentUsage
from waldur_api_client.models.component_user_usage_create_request import (
    ComponentUserUsageCreateRequest,
)
from waldur_api_client.models.component_user_usage_limit import ComponentUserUsageLimit
from waldur_api_client.models.marketplace_orders_list_state_item import (
    MarketplaceOrdersListStateItem,
)
from waldur_api_client.models.marketplace_provider_resources_list_state_item import (
    MarketplaceProviderResourcesListStateItem,
)
from waldur_api_client.models.offering_component import OfferingComponent
from waldur_api_client.models.offering_user import OfferingUser
from waldur_api_client.models.offering_user_state import OfferingUserState
from waldur_api_client.models.order_details import (
    OrderDetails,
)
from waldur_api_client.models.order_details_limits import (
    OrderDetailsLimits,
)
from waldur_api_client.models.order_set_state_erred_request import OrderSetStateErredRequest
from waldur_api_client.models.order_state import OrderState
from waldur_api_client.models.project_user import ProjectUser
from waldur_api_client.models.provider_offering_details import ProviderOfferingDetails
from waldur_api_client.models.request_types import RequestTypes
from waldur_api_client.models.resource import Resource as WaldurResource
from waldur_api_client.models.resource_backend_id_request import ResourceBackendIDRequest
from waldur_api_client.models.resource_backend_metadata_request import (
    ResourceBackendMetadataRequest,
)
from waldur_api_client.models.resource_state import ResourceState
from waldur_api_client.types import Unset

from waldur_site_agent.backend import BackendType, logger
from waldur_site_agent.backend import exceptions as backend_exceptions
from waldur_site_agent.backend import utils as backend_utils
from waldur_site_agent.backend.backends import BaseBackend
from waldur_site_agent.backend.exceptions import BackendError
from waldur_site_agent.backend.structures import BackendResourceInfo
from waldur_site_agent.common import agent_identity_management, pagination, structures, utils
from waldur_site_agent.common.structures import AccountType


class UsageAnomalyError(Exception):
    """Raised when usage anomaly is detected.

    This exception is raised when the system detects that new usage data
    is lower than previously reported usage for the same billing period,
    which typically indicates a data collection or processing error.
    """


class ObjectNotFoundError(Exception):
    """Raised when a required object cannot be found in Waldur or backend.

    This exception is used throughout the processors when attempting to
    retrieve resources, users, or other objects that should exist but
    cannot be located in either Waldur or the backend system.
    """


class OfferingBaseProcessor(abc.ABC):
    """Abstract base class for all offering processors.

    This class provides the common foundation for all offering processors,
    including Waldur API client setup, backend initialization, and shared
    utility methods. All concrete processor implementations inherit from
    this class to ensure consistent behavior and interface.

    The processor handles:
    - Waldur REST API client configuration and authentication
    - Backend system selection and initialization based on offering configuration
    - Username management backend setup
    - Common error handling and logging patterns

    Attributes:
        BACKEND_TYPE_KEY: Key used to identify which backend to use for this processor
        offering: The offering configuration being processed
        timezone: Timezone for billing period calculations
        waldur_rest_client: Authenticated client for Waldur API access
        resource_backend: Backend implementation for this offering
    """

    BACKEND_TYPE_KEY = "abstract"

    def __init__(
        self,
        offering: structures.Offering,
        waldur_rest_client: utils.AuthenticatedClient,
        timezone: str = "",
        resource_backend: Optional[BaseBackend] = None,
        resource_backend_version: Optional[str] = None,
    ) -> None:
        """Initialize the offering processor.

        Args:
            offering: The offering configuration to process
            waldur_rest_client: HTTP Client for Waldur API
            timezone: Timezone for billing period calculations (defaults to UTC)
            resource_backend: Backend instance for resource operations
                (optional, will be created if not provided)
            resource_backend_version: Version of the resource backend
                (optional, will be determined if not provided)

        Raises:
            BackendError: If unable to create a backend for the offering
        """
        self.offering: structures.Offering = offering
        self.timezone: str = timezone
        self.waldur_rest_client = waldur_rest_client

        # Use dependency injection if backend is provided, otherwise create it
        if resource_backend is not None:
            self.resource_backend = resource_backend
            self.resource_backend_version = resource_backend_version or "unknown"
        else:
            self.resource_backend, self.resource_backend_version = utils.get_backend_for_offering(
                offering, self.BACKEND_TYPE_KEY
            )

        if self.resource_backend.backend_type == BackendType.UNKNOWN.value:
            raise backend_exceptions.BackendError(f"Unable to create backend for {self.offering}")

        self._print_current_user()

        self.waldur_offering = marketplace_provider_offerings_retrieve.sync(
            client=self.waldur_rest_client, uuid=self.offering.uuid
        )
        utils.extend_backend_components(self.offering, self.waldur_offering.components)

        service_providers = marketplace_service_providers_list.sync(
            customer_uuid=self.waldur_offering.customer_uuid.hex,
            client=self.waldur_rest_client,
        )

        self.service_provider = service_providers[0]

    def _print_current_user(self) -> None:
        """Log information about the current authenticated Waldur user."""
        current_user = utils.get_current_user_from_client(self.waldur_rest_client)
        utils.print_current_user(current_user)

    @abc.abstractmethod
    def process_offering(self) -> None:
        """Process the offering according to the specific processor type.

        This method must be implemented by concrete processor classes to handle
        their specific responsibilities (order processing, membership sync, or reporting).
        """

    def _update_offering_users(self, offering_users: list[OfferingUser]) -> bool:
        """Generate usernames for offering users and update their state accordingly.

        Args:
            offering_users: List of offering users to process
        """
        return utils.update_offering_users(self.offering, self.waldur_rest_client, offering_users)

    def register(self, service: AgentService) -> AgentProcessor:
        """Register this processor in Waldur."""
        agent_identity_manager = agent_identity_management.AgentIdentityManager(
            self.offering, self.waldur_rest_client
        )
        processor_name = self.__class__.__name__
        backend_type = (
            self.resource_backend.__class__.__module__
            + "."
            + self.resource_backend.__class__.__name__
        )
        return agent_identity_manager.register_processor(
            service, processor_name, backend_type, self.resource_backend_version
        )


class OfferingOrderProcessor(OfferingBaseProcessor):
    """Processor for handling Waldur marketplace orders.

    This processor fetches pending and executing orders from Waldur and manages
    their lifecycle by creating, updating, or terminating backend resources.
    It handles the complete order workflow from approval through completion.

    The processor supports three types of orders:
    - CREATE: Creates new resources in the backend system
    - UPDATE: Modifies existing resource limits or configurations
    - TERMINATE: Removes resources from the backend system

    Order processing includes user context management, retry logic for failures,
    and comprehensive error handling with automatic order state updates.
    """

    BACKEND_TYPE_KEY = "order_processing_backend"

    def log_order_processing_error(self, order: OrderDetails, e: Exception) -> None:
        """Log detailed error information for order processing failures.

        Args:
            order: The order that failed to process
            e: The exception that occurred during processing
        """
        name = order.resource_name or "N/A"
        logger.exception(
            "Error while processing order %s (%s), type %s, state %s: %s",
            order.uuid,
            name,
            order.type_,
            order.state,
            e,
        )

    def process_offering(self) -> None:
        """Process all pending and executing orders for this offering.

        Fetches orders from Waldur with PENDING_PROVIDER or EXECUTING status
        and processes each order according to its type and current state.
        Includes error handling and logging for individual order failures.
        """
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )
        orders: list[OrderDetails] = marketplace_orders_list.sync(
            client=self.waldur_rest_client,
            offering_uuid=self.offering.uuid,
            state=[
                MarketplaceOrdersListStateItem.PENDING_PROVIDER,
                MarketplaceOrdersListStateItem.EXECUTING,
            ],
            page_size=100,
        )

        if not orders:
            logger.info("There are no pending or executing orders")
            return
        for order in orders:
            try:
                self.process_order_with_retries(order)
            except Exception as e:
                self.log_order_processing_error(order, e)

    def get_order_info(self, order_uuid: str) -> Optional[OrderDetails]:
        """Retrieve current order information from Waldur API.

        Args:
            order_uuid: UUID of the order to retrieve

        Returns:
            OrderDetails if found, None if retrieval fails
        """
        try:
            return marketplace_orders_retrieve.sync(client=self.waldur_rest_client, uuid=order_uuid)
        except UnexpectedStatus as e:
            logger.error("Failed to get order %s info: %s", order_uuid, e)
            return None

    def process_order_with_retries(
        self, order_info: OrderDetails, retry_count: int = 10, delay: int = 5
    ) -> None:
        """Process an order with automatic retry on failures.

        Args:
            order_info: The order to process
            retry_count: Maximum number of retry attempts (default: 10)
            delay: Delay in seconds between retry attempts (default: 5)
        """
        for attempt_number in range(retry_count):
            try:
                logger.info("Attempt %s of %s", attempt_number + 1, retry_count)
                order: Optional[OrderDetails] = self.get_order_info(order_info.uuid.hex)
                if order is None:
                    logger.error("Failed to get order %s info", order_info.uuid)
                    return
                self.process_order(order)
                break
            except UnexpectedStatus as e:
                if order is not None:
                    self.log_order_processing_error(order, e)
                else:
                    self.log_order_processing_error(order_info, e)
                logger.info("Retrying order %s processing in %s seconds", order_info.uuid, delay)
                sleep(delay)

        if attempt_number == retry_count - 1:
            logger.error(
                "Failed to process order %s after %s retries, skipping to the next one",
                order_info.uuid,
                retry_count,
            )

    def process_order(self, order: OrderDetails) -> None:
        """Process a single order through its complete lifecycle.

        Handles order approval (if needed), determines the order type,
        and delegates to the appropriate processing method. Updates
        order state to DONE on success or ERRED on failure.

        Args:
            order: The order to process
        """
        try:
            logger.info(
                "Processing order %s (%s) type %s, state %s",
                order.resource_name or "N/A",
                order.uuid,
                order.type_,
                order.state,
            )

            if order.state in [OrderState.DONE, OrderState.ERRED]:
                logger.info(
                    "Order %s (%s) is in a terminal state %s, skipping processing",
                    order.resource_name or "N/A",
                    order.uuid,
                    order.state,
                )
                return

            if order.state == OrderState.EXECUTING:
                logger.info("Order is executing already, no need for approval")
            elif order.state == OrderState.PENDING_PROVIDER:
                logger.info("Approving the order")
                marketplace_orders_approve_by_provider.sync_detailed(
                    client=self.waldur_rest_client, uuid=order.uuid.hex
                )
                logger.info("Refreshing the order")
                order = marketplace_orders_retrieve.sync(
                    client=self.waldur_rest_client, uuid=order.uuid.hex
                )
            else:
                logger.warning(
                    "The order %s %s (%s) is in unexpected state %s, skipping processing",
                    order.type_,
                    order.resource_name,
                    order.uuid,
                    order.state,
                )
                return

            order_is_done = False
            if order.type_ == RequestTypes.CREATE:
                order_is_done = self._process_create_order(order)

            if order.type_ == RequestTypes.UPDATE:
                order_is_done = self._process_update_order(order)

            if order.type_ == RequestTypes.TERMINATE:
                order_is_done = self._process_terminate_order(order)
            # TODO: no need for update of orders for marketplace SLURM offerings
            if order_is_done:
                logger.info("Marking order as done")
                marketplace_orders_set_state_done.sync_detailed(
                    client=self.waldur_rest_client, uuid=order.uuid
                )

                logger.info("The order has been successfully processed")

                # Refresh local order
                order = marketplace_orders_retrieve.sync(
                    client=self.waldur_rest_client, uuid=order.uuid.hex
                )

                self._post_process_order(order)
            else:
                logger.warning("The order processing was not finished, skipping to the next one")

        except Exception as e:
            logger.exception(
                "Error while processing order %s: %s",
                order.uuid,
                e,
            )
            if order.state != OrderState.DONE:
                order_set_state_erred_request = OrderSetStateErredRequest(
                    error_message=str(e),
                    error_traceback=traceback.format_exc(),
                )
                marketplace_orders_set_state_erred.sync_detailed(
                    client=self.waldur_rest_client,
                    uuid=order.uuid,
                    body=order_set_state_erred_request,
                )

    def _create_resource(
        self,
        waldur_resource: WaldurResource,
        user_context: dict,
    ) -> BackendResourceInfo | None:
        """Create a new resource in the backend system.

        Args:
            waldur_resource: Waldur resource information
            user_context: User context containing team and offering user data

        Returns:
            Created backend resource or None if creation failed

        Raises:
            BackendError: If resource creation fails
        """
        logger.info("Creating resource %s", waldur_resource.name)

        # Use the provided user context for resource creation
        backend_resource_info = self.resource_backend.create_resource(waldur_resource, user_context)
        if backend_resource_info.backend_id == "":
            msg = f"Unable to create a backend resource for offering {self.offering}"
            raise backend_exceptions.BackendError(msg)

        logger.info("Updating resource metadata in Waldur")
        marketplace_provider_resources_set_backend_id.sync(
            client=self.waldur_rest_client,
            uuid=waldur_resource.uuid.hex,
            body=ResourceBackendIDRequest(backend_id=backend_resource_info.backend_id),
        )
        logger.info("Resource backend id is set to %s", backend_resource_info.backend_id)

        if backend_resource_info.limits:
            marketplace_provider_resources_set_limits.sync(
                uuid=waldur_resource.uuid.hex,
                body=ResourceSetLimitsRequest(limits=backend_resource_info.limits),
                client=self.waldur_rest_client,
            )
            logger.info("Resource limits are set to %s", backend_resource_info.limits)

        return backend_resource_info

    def _fetch_user_context_for_resource(self, resource_uuid: str) -> dict:
        """Fetch comprehensive user context for resource operations.

        Retrieves project team members and offering users, creating mappings
        for efficient user lookup during resource creation and management.

        Args:
            resource_uuid: UUID of the resource to fetch context for

        Returns:
            Dictionary containing:
            - team: List of project team members
            - offering_users: List of offering users with usernames
            - user_mappings: Mapping of user UUIDs to ProjectUser objects
            - offering_user_mappings: Mapping of user UUIDs to OfferingUser objects

        Raises:
            ObjectNotFoundError: If no team members found for the resource
        """
        try:
            logger.info("Fetching user context for resource %s", resource_uuid)
            # Get project team members
            team: list[ProjectUser] | None = marketplace_provider_resources_team_list.sync(
                client=self.waldur_rest_client, uuid=resource_uuid
            )

            if not team:
                raise ObjectNotFoundError(f"No team members found for resource {resource_uuid}")  # noqa: TRY301

            user_uuids = {user.uuid for user in team}

            # Get offering users
            offering_users_all: list[OfferingUser] | None = pagination.get_all_paginated(
                marketplace_offering_users_list.sync,
                client=self.waldur_rest_client,
                offering_uuid=[self.offering.uuid],
                is_restricted=False,
            )

            if not offering_users_all:
                logger.warning("No offering users found for offering %s", self.offering.uuid)
                offering_users = []
            else:
                # Filter offering users to only those in the project team
                offering_users = [
                    offering_user
                    for offering_user in offering_users_all
                    if offering_user.user_uuid in user_uuids
                ]

            # Create user mappings for easy lookup
            user_mappings = {user.uuid: user for user in team}
            offering_user_mappings = {
                offering_user.user_uuid: offering_user for offering_user in offering_users
            }
        except Exception as e:
            logger.warning("Failed to fetch user context for resource %s: %s", resource_uuid, e)
            return {
                "team": [],
                "offering_users": [],
                "user_mappings": {},
                "offering_user_mappings": {},
            }
        else:
            return {
                "team": team,
                "offering_users": offering_users,
                "user_mappings": user_mappings,
                "offering_user_mappings": offering_user_mappings,
            }

    def _add_users_to_resource(
        self,
        waldur_resource: WaldurResource,
        user_context: dict,
    ) -> None:
        """Add offering users to the backend resource.

        Updates offering user usernames if needed and adds all users
        with valid usernames to the backend resource.

        Args:
            waldur_resource: The Waldur resource to add users to
            user_context: User context containing offering users and mappings
        """
        logger.info("Adding users to resource")
        offering_users = user_context["offering_users"]

        # Update offering user usernames (only for users with blank usernames)
        if self._update_offering_users(offering_users):
            # Refresh local user_context cache
            user_context_new = self._fetch_user_context_for_resource(waldur_resource.uuid.hex)
            user_context.update(user_context_new)

        logger.info(
            "Using %s (%s) offering users, user count: %s",
            self.offering.name,
            self.offering.uuid,
            len(user_context["offering_users"]),
        )

        # Filter only team members with usernames set and in OK state
        offering_usernames = {
            offering_user.username
            for offering_user in user_context["offering_users"]
            if offering_user.state == OfferingUserState.OK
        }

        if not offering_usernames:
            logger.info("No users to add to resource")
            return

        self.resource_backend.add_users_to_resource(
            waldur_resource.backend_id,
            offering_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

    def _process_create_order(self, order: OrderDetails | None) -> bool:
        """Process a CREATE order to establish a new resource.

        Waits for Waldur resource creation, fetches user context,
        creates the backend resource, and adds users to it.

        Args:
            order: The CREATE order to process

        Returns:
            True if processing completed successfully, False otherwise
        """
        if not order:
            logger.error("Error during order processing: Order is None")
            return False

        waldur_resource: WaldurResource | None = marketplace_provider_resources_retrieve.sync(
            uuid=order.marketplace_resource_uuid.hex, client=self.waldur_rest_client
        )
        if not waldur_resource:
            raise ObjectNotFoundError(
                f"Waldur resource {order.marketplace_resource_uuid.hex} not found"
            )

        create_resource = True
        if waldur_resource.backend_id != "":
            logger.info(
                "Waldur resource backend id is not empty %s, checking backend resource data",
                waldur_resource.backend_id,
            )
            backend_resource_info = self.resource_backend.pull_resource(waldur_resource)
            if backend_resource_info is not None:
                logger.info(
                    "Resource %s (%s) is already created, skipping creation",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                )
                create_resource = False

        # Fetch user context for resource creation
        user_context = self._fetch_user_context_for_resource(order.marketplace_resource_uuid.hex)

        backend_resource_info = None
        if create_resource:
            backend_resource_info = self._create_resource(waldur_resource, user_context)
            if backend_resource_info is None:
                msg = f"Unable to create the resource {waldur_resource.name}"
                raise backend_exceptions.BackendError(msg)

        if backend_resource_info is None:
            return False

        waldur_resource.backend_id = backend_resource_info.backend_id

        return True

    def _process_update_order(self, order: OrderDetails) -> bool:
        """Process an UPDATE order to modify resource limits.

        Args:
            order: The UPDATE order containing new limit specifications

        Returns:
            True if update completed successfully, False otherwise

        Raises:
            ObjectNotFoundError: If the target resource is not found
        """
        logger.info("Updating limits for %s", order.resource_name)
        resource_uuid = order.marketplace_resource_uuid
        waldur_resource: WaldurResource | None = marketplace_provider_resources_retrieve.sync(
            uuid=resource_uuid, client=self.waldur_rest_client
        )
        if not waldur_resource:
            raise ObjectNotFoundError(f"Waldur resource {resource_uuid} not found")

        waldur_resource_backend_id = waldur_resource.backend_id
        new_limits: Unset | OrderDetailsLimits = order.limits
        if not new_limits:
            logger.error(
                "Order %s (resource %s) with type" + "Update does not include new limits",
                order.uuid,
                waldur_resource.name,
            )

        self.resource_backend.set_resource_limits(waldur_resource_backend_id, new_limits.to_dict())

        logger.info(
            "The limits for %s were updated successfully from %s to %s",
            waldur_resource.name,
            order.resource_name or "N/A",
            new_limits.to_dict() if new_limits else "N/A",
        )
        return True

    def _process_terminate_order(self, order: OrderDetails) -> bool:
        """Process a TERMINATE order to remove a resource.

        Args:
            order: The TERMINATE order specifying the resource to remove

        Returns:
            True if termination completed successfully, False otherwise

        Raises:
            ObjectNotFoundError: If the target resource is not found
        """
        logger.info("Terminating resource %s", order.resource_name)
        resource_uuid = order.marketplace_resource_uuid
        waldur_resource: WaldurResource | None = marketplace_provider_resources_retrieve.sync(
            uuid=resource_uuid, client=self.waldur_rest_client
        )
        if not waldur_resource:
            raise ObjectNotFoundError(f"Waldur resource {resource_uuid} not found")
        project_slug = order.project_slug

        self.resource_backend.delete_resource(waldur_resource, project_slug=project_slug)

        logger.info("Allocation has been terminated successfully")
        return True

    def _post_process_order(self, order: OrderDetails) -> bool:
        if order.type_ == RequestTypes.CREATE:
            logger.info(
                "Running post process order script for %s (%s)",
                order.resource_name or "N/A",
                order.uuid.hex,
            )
            # Get Waldur resource data
            waldur_resource: WaldurResource = marketplace_provider_resources_retrieve.sync(
                uuid=order.marketplace_resource_uuid.hex, client=self.waldur_rest_client
            )

            # Fetch fresh user context for adding users
            user_context = self._fetch_user_context_for_resource(
                order.marketplace_resource_uuid.hex
            )

            # Add users to the backend resource
            self._add_users_to_resource(waldur_resource, user_context)

        return True


class OfferingMembershipProcessor(OfferingBaseProcessor):
    """Processor for synchronizing user memberships between Waldur and backends.

    This processor handles bidirectional synchronization of user access and
    memberships between Waldur and backend systems. It processes resource
    status changes, user additions/removals, and maintains consistency
    between the systems.

    Key responsibilities:
    - Synchronize user lists between Waldur project teams and backend resources
    - Handle resource status changes (paused, downscaled, restored)
    - Manage user limits and permissions on backend resources
    - Process event-driven user role changes
    - Update resource metadata and sync timestamps in Waldur

    The processor supports both full synchronization and incremental updates
    based on specific events or scheduled operations.
    """

    BACKEND_TYPE_KEY = "membership_sync_backend"

    def _get_waldur_resources(self, project_uuid: Optional[str] = None) -> list[WaldurResource]:
        """Fetch Waldur resources for this offering, optionally filtered by project.

        Args:
            project_uuid: If provided, only return resources from this project

        Returns:
            List of resources that have backend IDs and are in OK or ERRED state
        """
        filters: dict[str, list | str] = {
            "offering_uuid": [self.offering.uuid],
            "state": [
                MarketplaceProviderResourcesListStateItem.OK,
                MarketplaceProviderResourcesListStateItem.ERRED,
            ],
        }

        if project_uuid is not None:
            filters["project_uuid"] = project_uuid
        waldur_resources: list[WaldurResource] = pagination.get_all_paginated(
            marketplace_provider_resources_list.sync,
            client=self.waldur_rest_client,
            page_size=100,
            **filters,
        )

        waldur_resources_filtered = [
            waldur_resource for waldur_resource in waldur_resources if waldur_resource.backend_id
        ]

        logger.info(
            "Fetched %s resources (%s with backend_id set) under %s offering",
            len(waldur_resources),
            len(waldur_resources_filtered),
            self.offering.name,
        )

        return waldur_resources_filtered

    def process_resource_by_uuid(self, resource_uuid: str) -> None:
        """Process a specific resource's status and membership data.

        This method processes a single resource identified by UUID,
        performing full synchronization of its status and user memberships.

        Args:
            resource_uuid: UUID of the resource to process

        Raises:
            ObjectNotFoundError: If the resource is not found in Waldur
        """
        logger.info("Processing resource state and membership data, uuid: %s", resource_uuid)
        logger.info("Fetching resource from Waldur")
        waldur_resource: WaldurResource | None = marketplace_provider_resources_retrieve.sync(
            uuid=resource_uuid, client=self.waldur_rest_client
        )
        if not waldur_resource:
            raise ObjectNotFoundError(f"Waldur resource {resource_uuid} not found")
        logger.info(
            "Pulling resource %s (%s) from backend",
            waldur_resource.name,
            waldur_resource.backend_id,
        )
        resource_report = self.resource_backend.pull_resources([waldur_resource])
        if not resource_report:
            return
        self._process_resources(resource_report)

    def process_offering(self) -> None:
        """Process all resources in this offering for membership synchronization.

        Fetches all resources for the offering and processes each one to
        synchronize user memberships, resource status, and metadata between
        Waldur and the backend system.
        """
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )

        waldur_resources_info = self._get_waldur_resources()
        resource_report = self.resource_backend.pull_resources(waldur_resources_info)

        self._process_resources(resource_report)

    def _get_user_offering_users(
        self, user_uuid: str, offering_uuid: str | None = None
    ) -> list[OfferingUser]:
        """Fetch offering users for a specific user UUID.

        Args:
            user_uuid: UUID of the user to look up
            offering_uuid: optional UUID of the offering

        Returns:
            List of offering users associated with the specified user and the offering if specified

        Raises:
            ObjectNotFoundError: If no offering users found for the user
        """
        filters: dict[str, Any] = {"user_uuid": user_uuid, "is_restricted": False}
        if offering_uuid is not None:
            filters["offering_uuid"] = [offering_uuid]
        offering_users: list[OfferingUser] = marketplace_offering_users_list.sync(
            client=self.waldur_rest_client, **filters
        )
        return offering_users

    def process_user_role_changed(self, user_uuid: str, project_uuid: str, granted: bool) -> None:
        """Process a user role change event.

        Handles adding or removing a user from backend resources when their
        project role changes. Respects resource access restrictions.

        Args:
            user_uuid: UUID of the user whose role changed
            project_uuid: UUID of the project where the role changed
            granted: True if access was granted, False if revoked
        """
        offering_users: list[OfferingUser] = self._get_user_offering_users(
            user_uuid, self.offering.uuid
        )
        if len(offering_users) == 0:
            logger.info(
                "User %s is not linked to the offering %s (%s)",
                user_uuid,
                self.offering.name,
                self.offering.uuid,
            )
            return

        if self._update_offering_users(offering_users):
            # Refresh offering users after username generation
            offering_users = self._get_user_offering_users(user_uuid, self.offering.uuid)

        username = offering_users[0].username
        logger.info("Using offering user with username %s", username)
        if not username:
            logger.warning("Username is blank, skipping processing")
            return

        resources: list[WaldurResource] = self._get_waldur_resources(project_uuid=project_uuid)
        resource_report = self.resource_backend.pull_resources(resources)

        for waldur_resource, _ in resource_report.values():
            try:
                if granted:
                    if waldur_resource.restrict_member_access:
                        logger.info("The resource is restricted, skipping new role.")
                        continue
                    self.resource_backend.add_user(waldur_resource.backend_id, username)
                else:
                    self.resource_backend.remove_user(waldur_resource.backend_id, username)
            except Exception as exc:
                logger.error(
                    "Unable to add user %s to the resource %s, error: %s",
                    username,
                    waldur_resource.backend_id,
                    exc,
                )

    def _refresh_local_offering_users(self) -> list[OfferingUser]:
        try:
            # Fetch offering users
            offering_users: list[OfferingUser] = self._get_waldur_offering_users()

            # Update offering user usernames
            if self._update_offering_users(offering_users):
                # Refresh offering users after username generation
                return self._get_waldur_offering_users()
            return offering_users
        except Exception as exc:
            logger.error("Unable to refresh local offering users: %s", exc)
            return []

    def process_project_user_sync(self, project_uuid: str) -> None:
        """Perform full user synchronization for all resources in a project.

        This method synchronizes all users across all resources within
        a specific project, ensuring consistency between Waldur project
        teams and backend resource access.

        Args:
            project_uuid: UUID of the project to synchronize
        """
        logger.info("Processing sync of all users for project %s", project_uuid)
        resources = self._get_waldur_resources(project_uuid=project_uuid)
        resource_report = self.resource_backend.pull_resources(resources)
        # Fetch offering users
        offering_users = self._refresh_local_offering_users()

        for waldur_resource, backend_resource_info in resource_report.values():
            try:
                logger.info(
                    "Syncing users for resource %s (%s)",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                )
                self._sync_resource_users(waldur_resource, backend_resource_info, offering_users)
            except Exception as exc:
                logger.error(
                    "Unable to sync resource %s (%s), error: %s",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                    exc,
                )

    def _get_waldur_offering_users(self) -> list[OfferingUser]:
        """Fetch all offering users for this offering.

        Returns:
            List of all non-restricted offering users for this offering

        Raises:
            ObjectNotFoundError: If no offering users found
        """
        logger.info("Fetching Waldur offering users (OK and Requested)")
        offering_users: list[OfferingUser] = pagination.get_all_paginated(
            marketplace_offering_users_list.sync,
            client=self.waldur_rest_client,
            offering_uuid=[self.offering.uuid],
            state=[
                MarketplaceOfferingUsersListStateItem.OK,
                MarketplaceOfferingUsersListStateItem.REQUESTED,
            ],
            is_restricted=False,
        )
        logger.info("Fetched %d offering users", len(offering_users))
        return offering_users

    def _get_waldur_resource_team(self, resource: WaldurResource) -> list[ProjectUser]:
        logger.info("Fetching Waldur resource team")
        return marketplace_provider_resources_team_list.sync(
            client=self.waldur_rest_client, uuid=resource.uuid.hex
        )

    def _group_resource_usernames(
        self,
        waldur_resource: WaldurResource,
        backend_resource_info: BackendResourceInfo,
        offering_users: list[OfferingUser],
    ) -> tuple[set[str], set[str], set[str]]:
        logger.info("Fetching new, existing and stale resource users")
        usernames: list[str] = backend_resource_info.users
        local_usernames = set(usernames)
        logger.info(
            "The usernames from the backend (%s): %s",
            len(local_usernames),
            ", ".join(local_usernames),
        )

        # Offering users sync
        # The service fetches offering users from Waldur and pushes them to the cluster
        # If an offering user is not in the team anymore, it will be removed from the backend
        team: list[ProjectUser] = self._get_waldur_resource_team(waldur_resource)

        offering_user_usernames = {
            offering_user.username for offering_user in offering_users if offering_user.username
        }

        resource_usernames = {
            user.offering_user_username for user in team if user.offering_user_username
        }
        logger.info(
            "Resource offering usernames (%s): %s",
            len(resource_usernames),
            ", ".join(str(u) for u in resource_usernames),
        )

        logger.info("Number of offering user usernames: %s", len(offering_user_usernames))

        existing_usernames: set[str] = resource_usernames & local_usernames
        logger.info(
            "Resource existing usernames (%s): %s",
            len(existing_usernames),
            ", ".join(existing_usernames),
        )

        new_usernames: set[str] = resource_usernames - local_usernames
        logger.info("Resource new usernames (%s): %s", len(new_usernames), ", ".join(new_usernames))

        stale_usernames: set[str] = (offering_user_usernames & local_usernames) - resource_usernames
        logger.info(
            "Resource stale usernames (%s): %s", len(stale_usernames), ", ".join(stale_usernames)
        )

        return existing_usernames, stale_usernames, new_usernames

    def _sync_resource_users(
        self,
        waldur_resource: WaldurResource,
        backend_resource_info: BackendResourceInfo,
        offering_users: list[OfferingUser],
    ) -> set[str]:
        """Sync users for the resource between Waldur and the site.

        return: the actual resource usernames (existing + added)
        """
        logger.info("Syncing user list for resource %s", waldur_resource.name)

        existing_usernames, stale_usernames, new_usernames = self._group_resource_usernames(
            waldur_resource, backend_resource_info, offering_users
        )

        if waldur_resource.restrict_member_access:
            # The idea is to remove the existing associations in both sides
            # and avoid creation of new associations
            logger.info(
                "Resource is restricted for members, removing all the existing associations"
            )

            self.resource_backend.remove_users_from_resource(
                waldur_resource.backend_id, existing_usernames
            )
            return set()

        added_usernames = self.resource_backend.add_users_to_resource(
            waldur_resource.backend_id,
            new_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

        self.resource_backend.remove_users_from_resource(
            waldur_resource.backend_id,
            stale_usernames,
        )

        return existing_usernames | added_usernames

    def _sync_resource_status(self, waldur_resource: WaldurResource) -> None:
        """Syncs resource status between Waldur and the backend."""
        logger.info(
            "Syncing resource status for resource %s (%s)",
            waldur_resource.name,
            waldur_resource.backend_id,
        )
        if waldur_resource.paused:
            logger.info("Resource pausing is requested, processing it")
            pausing_done = self.resource_backend.pause_resource(waldur_resource.backend_id)
            if pausing_done:
                logger.info("Pausing is successfully completed")
            else:
                logger.warning("Pausing is not done")
        elif waldur_resource.downscaled:
            logger.info("Resource downscaling is requested, processing it")
            downscaling_done = self.resource_backend.downscale_resource(waldur_resource.backend_id)
            if downscaling_done:
                logger.info("Downscaling is successfully completed")
            else:
                logger.warning("Downscaling is not done")
        else:
            logger.info(
                "The resource is not downscaled or paused, resetting the QoS to the default one"
            )
            restoring_done = self.resource_backend.restore_resource(waldur_resource.backend_id)
            if restoring_done:
                logger.info("Restoring is successfully completed")
            else:
                logger.info("Restoring is skipped")

        resource_metadata = self.resource_backend.get_resource_metadata(waldur_resource.backend_id)
        marketplace_provider_resources_set_backend_metadata.sync(
            uuid=waldur_resource.uuid.hex,
            client=self.waldur_rest_client,
            body=ResourceBackendMetadataRequest(backend_metadata=resource_metadata),
        )

    def _sync_resource_limits(self, waldur_resource: WaldurResource) -> None:
        """Syncs resource limits between Waldur and the backend."""
        utils.sync_waldur_resource_limits(
            self.resource_backend, self.waldur_rest_client, waldur_resource
        )

    # TODO: adapt for RabbitMQ-based processing
    # introduce new event and add support for the event in the agent
    def _sync_resource_user_limits(
        self, waldur_resource: WaldurResource, usernames: set[str]
    ) -> None:
        logger.info(
            "Synching resource user limits for resource %s (%s)",
            waldur_resource.name,
            waldur_resource.backend_id,
        )
        if waldur_resource.restrict_member_access:
            logger.info("Resource is restricted for members, skipping user limits setup")
            return

        backend_user_limits = self.resource_backend.get_resource_user_limits(
            waldur_resource.backend_id
        )

        for username in usernames:
            try:
                logger.info(
                    "Fetching user usage limits for %s, resource %s",
                    username,
                    waldur_resource.uuid.hex,
                )
                user_limits: list[ComponentUserUsageLimit] = component_user_usage_limits_list.sync(
                    client=self.waldur_rest_client,
                    resource_uuid=waldur_resource.uuid.hex,
                    username=username,
                )
                if len(user_limits) == 0:
                    existing_user_limits = backend_user_limits.get(username)
                    logger.debug("The limits for user %s are not set in Waldur", username)
                    if not existing_user_limits:
                        continue
                    logger.info("Unsetting the existing limits %s", existing_user_limits)
                    user_component_limits = {}
                else:
                    user_component_limits = {
                        user_limit.component_type: int(float(user_limit.limit))
                        for user_limit in user_limits
                    }
                self.resource_backend.set_resource_user_limits(
                    waldur_resource.backend_id, username, user_component_limits
                )
            except Exception as exc:
                logger.error(
                    "Unable to set user %s limits for resource %s (%s), reason: %s",
                    username,
                    waldur_resource.name,
                    waldur_resource.backend_id,
                    exc,
                )

    def _sync_resource_service_accounts(self, waldur_resource: WaldurResource) -> None:
        """Sync project service accounts between Waldur and the backend resource."""
        logger.info(
            "Syncing service accounts for the resource %s (%s)",
            waldur_resource.name,
            waldur_resource.backend_id,
        )
        if self.service_provider is None:
            logger.warning("No service provider configured, skipping service accounts sync")
            return

        service_accounts = pagination.get_all_paginated(
            marketplace_service_providers_project_service_accounts_list.sync,
            service_provider_uuid=self.service_provider.uuid.hex,
            project_uuid=waldur_resource.project_uuid.hex,
            client=self.waldur_rest_client,
        )
        usernames_active = {
            account.username
            for account in service_accounts
            if account.username and account.state == ServiceAccountState.OK
        }
        self.resource_backend.add_users_to_resource(waldur_resource.backend_id, usernames_active)

        usernames_closed = {
            account.username
            for account in service_accounts
            if account.username and account.state == ServiceAccountState.CLOSED
        }
        self.resource_backend.remove_users_from_resource(
            waldur_resource.backend_id, usernames_closed
        )

    def _sync_resource_course_accounts(self, waldur_resource: WaldurResource) -> None:
        """Sync course accounts between Waldur and the backend resource."""
        logger.info(
            "Syncing course accounts for the resource %s (%s)",
            waldur_resource.name,
            waldur_resource.backend_id,
        )
        if self.service_provider is None:
            logger.warning("No service provider configured, skipping service accounts sync")
            return

        course_accounts = pagination.get_all_paginated(
            marketplace_service_providers_course_accounts_list.sync,
            service_provider_uuid=self.service_provider.uuid.hex,
            project_uuid=waldur_resource.project_uuid.hex,
            client=self.waldur_rest_client,
        )
        usernames_active = {
            account.username
            for account in course_accounts
            if account.username and account.state == ServiceAccountState.OK
        }
        self.resource_backend.add_users_to_resource(waldur_resource.backend_id, usernames_active)

        usernames_closed = {
            account.username
            for account in course_accounts
            if account.username and account.state == ServiceAccountState.CLOSED
        }
        self.resource_backend.remove_users_from_resource(
            waldur_resource.backend_id, usernames_closed
        )

    def _process_resources(
        self,
        resource_report: dict[str, tuple[WaldurResource, BackendResourceInfo]],
    ) -> None:
        """Sync status and membership data for the resource."""
        # Fetch offering users
        offering_users = self._refresh_local_offering_users()

        for waldur_resource, backend_resource_info in resource_report.values():
            try:
                resource_usernames = self._sync_resource_users(
                    waldur_resource, backend_resource_info, offering_users
                )
                self._sync_resource_service_accounts(waldur_resource)
                self._sync_resource_course_accounts(waldur_resource)
                self._sync_resource_status(waldur_resource)
                self._sync_resource_limits(waldur_resource)
                self._sync_resource_user_limits(waldur_resource, resource_usernames)

                logger.info(
                    "Refreshing resource %s (%s) last sync",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                )

                marketplace_provider_resources_refresh_last_sync.sync_detailed(
                    uuid=waldur_resource.uuid.hex,
                    client=self.waldur_rest_client,
                )
                if waldur_resource.state == ResourceState.ERRED:
                    logger.info(
                        "Setting resource %s (%s) state to OK",
                        waldur_resource.name,
                        waldur_resource.backend_id,
                    )
                    marketplace_provider_resources_set_as_ok.sync_detailed(
                        uuid=waldur_resource.uuid.hex,
                        client=self.waldur_rest_client,
                    )
            except Exception as e:
                logger.exception(
                    "Error while processing allocation %s: %s",
                    waldur_resource.backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [waldur_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )

    def process_account_creation(self, account_username: str, account_type: AccountType) -> None:
        """Process service or course account creation."""
        params = {
            "service_provider_uuid": self.service_provider.uuid.hex,
            "username": account_username,
            "client": self.waldur_rest_client,
        }
        if account_type == AccountType.SERVICE_ACCOUNT:
            accounts = marketplace_service_providers_project_service_accounts_list.sync(**params)
        else:
            accounts = marketplace_service_providers_course_accounts_list.sync(**params)

        if len(accounts) == 0:
            logger.info(
                "No %s accounts found with username %s, skipping processing for offering %s (%s)",
                account_type.value,
                account_username,
                self.offering.uuid,
                self.offering.name,
            )
            return

        account: ProjectServiceAccount | CourseAccount = accounts[0]
        resources = self._get_waldur_resources(account.project_uuid.hex)
        for resource in resources:
            try:
                self.resource_backend.add_users_to_resource(resource.backend_id, {account_username})
            except BackendError as e:
                logger.error(
                    "Unable to add the account %s to resource %s, reason: %s",
                    account_username,
                    resource.backend_id,
                    e,
                )

    def process_account_removal(self, account_username: str, project_uuid: str) -> None:
        """Process course account removal."""
        resources = self._get_waldur_resources(project_uuid)
        for resource in resources:
            try:
                self.resource_backend.remove_users_from_resource(
                    resource.backend_id, {account_username}
                )
            except BackendError as e:
                logger.error(
                    "Unable to remove the account %s from resource %s, reason: %s",
                    account_username,
                    resource.backend_id,
                    e,
                )


class OfferingReportProcessor(OfferingBaseProcessor):
    """Processor for collecting and reporting usage data from backends to Waldur.

    This processor handles the collection of resource usage data from backend
    systems and reports it to Waldur for billing and monitoring purposes.
    It processes both total resource usage and per-user usage breakdowns.

    Key responsibilities:
    - Collect usage data from backend systems for all resources
    - Validate usage data for anomalies (decreasing usage patterns)
    - Report total resource usage to Waldur marketplace
    - Report per-user usage breakdowns for detailed billing
    - Handle error cases and mark resources as erred when backend data is missing
    - Implement retry logic for transient failures

    The processor includes anomaly detection to prevent reporting usage data
    that appears to have decreased from previous reports, which typically
    indicates a data collection error.
    """

    BACKEND_TYPE_KEY = "reporting_backend"

    def process_offering(self) -> None:
        """Process all resources in this offering for usage reporting.

        Fetches all OK and ERRED resources for the offering and processes
        each one to collect and report usage data to Waldur. Includes
        error handling for individual resource failures.
        """
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )
        waldur_offering: ProviderOfferingDetails | None = (
            marketplace_provider_offerings_retrieve.sync(
                client=self.waldur_rest_client, uuid=self.offering.uuid
            )
        )
        waldur_resources = pagination.get_all_paginated(
            marketplace_provider_resources_list.sync,
            client=self.waldur_rest_client,
            offering_uuid=[self.offering.uuid],
            state=[
                MarketplaceProviderResourcesListStateItem.OK,
                MarketplaceProviderResourcesListStateItem.ERRED,
            ],
        )
        if not waldur_resources:
            logger.info("No resources to process")
            return

        logger.info(
            "Fetched %s resources under %s offering", len(waldur_resources), self.offering.name
        )

        for waldur_resource in waldur_resources:
            try:
                self._process_resource_with_retries(waldur_resource, waldur_offering)
            except Exception as e:
                logger.exception(
                    "Error while processing allocation %s: %s",
                    waldur_resource.backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [waldur_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )

    def _process_resource_with_retries(
        self,
        waldur_resource: WaldurResource,
        waldur_offering: ProviderOfferingDetails,
        retry_count: int = 10,
        delay: int = 5,
    ) -> None:
        for attempt_number in range(retry_count):
            try:
                logger.info(
                    "Attempt %s of %s, processing resource usage %s (%s)",
                    attempt_number + 1,
                    retry_count,
                    waldur_resource.name,
                    waldur_resource.backend_id,
                )
                self._process_resource(waldur_resource, waldur_offering)
                break
            except Exception as e:
                logger.warning(
                    "Error while processing resource %s (%s): %s",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                    e,
                )
                logger.info(
                    "Retrying resource usage %s processing in %s seconds",
                    waldur_resource.backend_id,
                    delay,
                )
                if attempt_number == retry_count - 1:
                    # If last attempt failed, raise the exception
                    logger.warning(
                        "Failed to process resource usage %s after %s retries,"
                        "skipping to the next resource",
                        waldur_resource.backend_id,
                        retry_count,
                    )
                    raise
                # If not last attempt, wait and retry
                sleep(delay)

    def _check_usage_anomaly(
        self,
        component_type: str,
        current_usage: float,
        existing_usages: list[ComponentUsage] | None,
    ) -> bool:
        """Check if the current usage is lower than existing usage."""
        if not existing_usages:
            return False
        # Find all usage records for this component type
        component_usages = [usage for usage in existing_usages if usage.type_ == component_type]

        if not component_usages:
            return False

        if len(component_usages) > 1:
            logger.error(
                "Found multiple usage records for component %s in the same billing period: %d",
                component_type,
                len(component_usages),
            )
            return True

        component_usage = component_usages[0]
        existing_usage = float(component_usage.usage)

        if current_usage < existing_usage:
            logger.error(
                "Usage anomaly detected for component %s: "
                "Current usage %s is lower than existing usage %s",
                component_type,
                current_usage,
                existing_usage,
            )
            return True

        return False

    def _submit_total_usage_for_resource(
        self,
        waldur_resource: WaldurResource,
        total_usage: dict[str, float],
        waldur_components: list[OfferingComponent],
    ) -> None:
        """Reports total usage for a backend resource to Waldur."""
        logger.info("Setting usages for %s: %s", waldur_resource.backend_id, total_usage)
        resource_uuid = waldur_resource.uuid.hex

        component_types: list[str] = [
            component.type_ for component in waldur_components if component.type_
        ]
        missing_components = set(total_usage) - set(component_types)
        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )

        current_time = backend_utils.get_current_time_in_timezone(self.timezone)
        month_start = backend_utils.month_start(current_time).date()
        existing_usages: list[ComponentUsage] = marketplace_component_usages_list.sync(
            client=self.waldur_rest_client, resource_uuid=resource_uuid, billing_period=month_start
        )
        for component, amount in total_usage.items():
            if component in component_types and self._check_usage_anomaly(
                component, amount, existing_usages
            ):
                logger.warning(
                    "Skipping usage update for resource %s due to anomaly detection",
                    waldur_resource.backend_id,
                )
                raise UsageAnomalyError(f"Usage anomaly detected for component {component}")

        usage_objects = [
            ComponentUsageItemRequest(type_=component, amount=str(amount))
            for component, amount in total_usage.items()
            if component in component_types
        ]
        request_body = ComponentUsageCreateRequest(usages=usage_objects, resource=resource_uuid)
        marketplace_component_usages_set_usage.sync_detailed(
            client=self.waldur_rest_client, body=request_body
        )

    def _submit_user_usage_for_resource(
        self,
        username: str,
        user_usage: dict[str, float],
        waldur_component_usages: list[ComponentUsage] | None,
    ) -> None:
        """Reports per-user usage for a backend resource to Waldur."""
        logger.info("Setting usages for %s", username)
        if not waldur_component_usages:
            logger.warning(
                "No component usages found for resource %s",
                username,
            )
            return
        component_usage_types = [
            component_usage.type_
            for component_usage in waldur_component_usages
            if component_usage.type_
        ]
        missing_components = set(user_usage) - set(component_usage_types)

        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )
        # Assumed to be looking up offering users by the user's username
        offering_users: list[OfferingUser] = marketplace_offering_users_list.sync(
            client=self.waldur_rest_client, user_username=username, query=self.offering.uuid
        )
        offering_user = offering_users[0] if offering_users else None

        for component_usage in waldur_component_usages:
            component_type = component_usage.type_
            usage = user_usage.get(component_type)
            if usage is None:
                logger.warning(
                    "No usage for Waldur component %s is found in SLURM user usage report",
                    component_type,
                )
                continue
            logger.info(
                "Submitting usage for username %s: %s -> %s",
                username,
                component_type,
                usage,
            )
            body = ComponentUserUsageCreateRequest(
                username=username,
                usage=usage,
            )
            if offering_user:
                body.user = offering_user.url
            else:
                logger.warning("No offering user found for username %s", username)

            marketplace_component_usages_set_user_usage.sync_detailed(
                uuid=component_usage.uuid, client=self.waldur_rest_client, body=body
            )

    def _process_resource(
        self,
        waldur_resource: WaldurResource,
        waldur_offering: ProviderOfferingDetails,
    ) -> None:
        """Processes usage report for the resource."""
        current_time = backend_utils.get_current_time_in_timezone(self.timezone)
        month_start = backend_utils.month_start(current_time).date()
        resource_backend_id = waldur_resource.backend_id
        logger.info("Pulling resource %s (%s)", waldur_resource.name, resource_backend_id)
        backend_resource_info = self.resource_backend.pull_resource(waldur_resource)
        if backend_resource_info is None:
            logger.info("The resource %s is missing in backend", resource_backend_id)
            if waldur_resource.state != ResourceState.ERRED:
                logger.info("Marking resource %s as erred in Waldur", resource_backend_id)
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [waldur_resource],
                    {
                        "error_message": f"The resource {resource_backend_id} "
                        "is missing on the backend"
                    },
                )
            return

        # Invalidate cache of Waldur resource
        logger.info(
            "Fetching Waldur resource data for %s (%s)", waldur_resource.name, resource_backend_id
        )
        waldur_resource_info: WaldurResource = marketplace_provider_resources_retrieve.sync(
            client=self.waldur_rest_client, uuid=waldur_resource.uuid.hex
        )

        if waldur_resource_info.state not in [
            MarketplaceProviderResourcesListStateItem.OK,
            MarketplaceProviderResourcesListStateItem.ERRED,
        ]:
            logger.error(
                "Waldur resource %s (%s) has incorrect state %s, skipping processing",
                waldur_resource_info.name,
                waldur_resource_info.backend_id,
                waldur_resource_info.state,
            )
            return
        # Set resource state OK if it is erred
        if waldur_resource_info.state == ResourceState.ERRED:
            marketplace_provider_resources_set_as_ok.sync_detailed(
                uuid=waldur_resource_info.uuid,
                client=self.waldur_rest_client,
            )

        usages: dict[str, dict[str, float]] = backend_resource_info.usage

        # Submit usage
        total_usage = usages.pop("TOTAL_ACCOUNT_USAGE")
        try:
            self._submit_total_usage_for_resource(
                waldur_resource_info,
                total_usage,
                waldur_offering.components,
            )
        except UsageAnomalyError:
            logger.info("Skipping per-user usage processing due to anomaly in usage reporting")
            return

        # Skip the following actions if the dict is empty
        if not usages:
            return

        waldur_component_usages: list[ComponentUsage] | None = (
            marketplace_component_usages_list.sync(
                client=self.waldur_rest_client,
                resource_uuid=waldur_resource_info.uuid.hex,
                date_after=month_start,
            )
        )
        logger.info("Setting per-user usages")
        for username, user_usage in usages.items():
            self._submit_user_usage_for_resource(username, user_usage, waldur_component_usages)


class OfferingImportableResourcesProcessor(OfferingBaseProcessor):
    """Processes importable resources for the offering and reports them to Waldur."""

    BACKEND_TYPE_KEY = "order_processing_backend"

    def process_offering(self) -> None:
        """This function is blank because the processor operates over backend resource request."""

    def _process_importable_resource(self, local_resource_info: BackendResourceInfo) -> None:
        logger.info("Processing importable resource %s", local_resource_info.backend_id)
        project_prefix = self.offering.backend_settings.get("project_prefix", "")
        parent_id = local_resource_info.parent_id
        if not parent_id.startswith(project_prefix):
            logger.info(
                "The parent_id %s of the resource %s does not have a required prefix, skipping it",
                parent_id,
                local_resource_info.backend_id,
            )
            return

        project_slug = parent_id.removeprefix(project_prefix)
        waldur_projects = projects_list.sync(
            slug=project_slug,
            client=self.waldur_rest_client,
        )
        if not waldur_projects:
            logger.info(
                "No Waldur project found for slug %s, skipping resource %s import",
                project_slug,
                local_resource_info.backend_id,
            )
            return
        waldur_project = waldur_projects[0]
        existing_backend_resources = pagination.get_all_paginated(
            backend_resources_list.sync,
            backend_id=local_resource_info.backend_id,
            project_uuid=waldur_project.uuid,
            offering_uuid=self.offering.uuid,
            client=self.waldur_rest_client,
        )

        if existing_backend_resources:
            logger.info(
                "Backend resource with id %s already exists in Waldur project %s, "
                "skipping submission",
                local_resource_info.backend_id,
                waldur_project.uuid.hex,
            )
            return

        logger.info(
            "Submitting backend resource %s to Waldur project %s",
            local_resource_info.backend_id,
            waldur_project.uuid.hex,
        )
        limits = self.resource_backend.get_resource_limits(local_resource_info.backend_id)
        backend_metadata = {
            "limits": limits,
        }
        payload = BackendResourceRequest(
            name=local_resource_info.backend_id,
            project=waldur_project.uuid,
            offering=self.offering.uuid,
            backend_id=local_resource_info.backend_id,
            backend_metadata=backend_metadata,
        )
        backend_resources_create.sync(
            body=payload,
            client=self.waldur_rest_client,
        )

    def _pre_process_backend_resource_request(self, request_uuid: str) -> None:
        backend_resource_request = backend_resource_requests_retrieve.sync(
            uuid=request_uuid,
            client=self.waldur_rest_client,
        )
        backend_resource_requests_start_processing.sync(
            uuid=backend_resource_request.uuid.hex,
            client=self.waldur_rest_client,
        )

    def _get_waldur_resources(self) -> dict[str, WaldurResource]:
        waldur_resources = pagination.get_all_paginated(
            marketplace_provider_resources_list.sync,
            offering_uuid=[self.offering.uuid],
            state=[
                MarketplaceProviderResourcesListStateItem.OK,
                MarketplaceProviderResourcesListStateItem.ERRED,
                MarketplaceProviderResourcesListStateItem.CREATING,
            ],
            client=self.waldur_rest_client,
        )

        waldur_resources_filtered = {
            resource.backend_id: resource for resource in waldur_resources if resource.backend_id
        }

        logger.info(
            "Fetched %s resources (%s with backend_id set) under %s offering",
            len(waldur_resources),
            len(waldur_resources_filtered),
            self.offering.name,
        )
        return waldur_resources_filtered

    def process_request(self, request_uuid: str) -> None:
        """Process backend resource request.

        List all resource in the backend, compare their backend_ids with ones from Waldur
        and report only ones absent in Waldur.
        """
        try:
            logger.info(
                "Processing backend resource request %s for offering %s",
                request_uuid,
                self.offering.name,
            )

            self._pre_process_backend_resource_request(request_uuid)
            waldur_resources: dict[str, WaldurResource] = self._get_waldur_resources()

            local_resource_list = self.resource_backend.list_resources()
            local_resources = {resource.backend_id: resource for resource in local_resource_list}

            importable_resource_backend_ids = local_resources.keys() - waldur_resources.keys()
            logger.info(
                "Found %d importable resources in the backend",
                len(importable_resource_backend_ids),
            )

            for backend_id in importable_resource_backend_ids:
                try:
                    local_resource = local_resources[backend_id]
                    self._process_importable_resource(local_resource)
                except Exception as e:
                    logger.error(
                        "Unable to process importable resource %s reason: %s", backend_id, e
                    )

            logger.info("Setting backend resource request %s as done", request_uuid)
            backend_resource_requests_set_done.sync(
                uuid=request_uuid,
                client=self.waldur_rest_client,
            )
        except Exception as e:
            logger.info("Unable to process importable resources reason: %s", e)
            payload = BackendResourceRequestSetErredRequest(
                error_message=str(e),
                error_traceback=traceback.format_exc(),
            )
            backend_resource_requests_set_erred.sync(
                uuid=request_uuid,
                client=self.waldur_rest_client,
                body=payload,
            )



"""Functions for management of MQTT signals."""

from __future__ import annotations

import signal
import sys
import types
from collections.abc import Generator
from contextlib import contextmanager
from typing import Union

import httpx
import paho.mqtt.client as mqtt
from waldur_api_client import errors
from waldur_api_client.api.marketplace_orders import marketplace_orders_list
from waldur_api_client.models.observable_object_type_enum import ObservableObjectTypeEnum

from waldur_site_agent.backend import logger
from waldur_site_agent.common import agent_identity_management
from waldur_site_agent.common import processors as common_processors
from waldur_site_agent.common import structures as common_structures
from waldur_site_agent.common.utils import get_client
from waldur_site_agent.event_processing import handlers
from waldur_site_agent.event_processing.event_subscription_manager import EventSubscriptionManager
from waldur_site_agent.event_processing.structures import (
    MqttConsumer,
    MqttConsumersMap,
    StompConsumer,
    StompConsumersMap,
    UserData,
)


def on_connect(
    client: mqtt.Client,
    userdata: UserData,
    flags: mqtt.ConnectFlags,
    reason_code: mqtt.ReasonCode,
    properties: Union[mqtt.Properties, None],
) -> None:
    """Order-processing handler for MQTT connection event."""
    del flags, properties
    logger.debug("Consumer connected with result code %s", reason_code)
    offering_uuid = userdata["offering"].uuid
    if reason_code.is_failure:
        logger.error("Consumer connection error (%s): %s", offering_uuid, reason_code.getName())
    else:
        event_subscription_uuid = userdata["event_subscription"]["uuid"]
        topic_postfix = userdata["topic_postfix"]
        topic_name = (
            f"subscription/{event_subscription_uuid}/offering/{offering_uuid}/{topic_postfix}"
        )
        logger.debug("Subscribing to the topic %s", topic_name)
        client.subscribe(topic_name)


def setup_stomp_offering_subscriptions(
    waldur_offering: common_structures.Offering, waldur_user_agent: str, global_proxy: str = ""
) -> list[StompConsumer]:
    """Set up STOMP subscriptions for the specified offering."""
    stomp_connections: list[StompConsumer] = []
    object_types = []

    if waldur_offering.order_processing_backend:
        object_types.append(ObservableObjectTypeEnum.ORDER)
    else:
        logger.info(
            "Order processing is disabled for offering %s, skipping start of STOMP connections",
            waldur_offering.name,
        )

    if waldur_offering.membership_sync_backend:
        object_types.extend(
            [
                ObservableObjectTypeEnum.USER_ROLE,
                ObservableObjectTypeEnum.RESOURCE,
                ObservableObjectTypeEnum.SERVICE_ACCOUNT,
                ObservableObjectTypeEnum.COURSE_ACCOUNT,
            ]
        )
    else:
        logger.info(
            "Membership sync is disabled for offering %s, skipping start of STOMP connections",
            waldur_offering.name,
        )

    if waldur_offering.resource_import_enabled:
        object_types.append(ObservableObjectTypeEnum.IMPORTABLE_RESOURCES)
    else:
        logger.info(
            "Resource import is disabled for offering %s, skipping start of STOMP connections",
            waldur_offering.name,
        )

    waldur_rest_client = get_client(
        waldur_offering.api_url,
        waldur_offering.api_token,
        waldur_user_agent,
        verify_ssl=waldur_offering.verify_ssl,
        proxy=global_proxy,
    )

    agent_identity_manager = agent_identity_management.AgentIdentityManager(
        waldur_offering, waldur_rest_client
    )
    identity_name = f"agent-{waldur_offering.uuid}"
    try:
        agent_identity = agent_identity_manager.register_identity(identity_name)
    except (errors.UnexpectedStatus, httpx.TimeoutException) as e:
        logger.exception(
            "Failed to register identity for the offering %s: %s", waldur_offering.name, e
        )
        return stomp_connections

    for object_type in object_types:
        try:
            event_subscription = agent_identity_manager.register_event_subscription(
                agent_identity, object_type
            )
            event_subscription_manager = EventSubscriptionManager(
                waldur_offering, None, None, waldur_user_agent, object_type, global_proxy
            )
            connection = event_subscription_manager.setup_stomp_connection(
                event_subscription,
                waldur_offering.stomp_ws_host,
                waldur_offering.stomp_ws_port,
                waldur_offering.stomp_ws_path,
            )
            connected = event_subscription_manager.start_stomp_connection(
                event_subscription, connection
            )
            if not connected:
                logger.error(
                    "Failed to start STOMP connection for the offering %s (%s), object type %s",
                    waldur_offering.name,
                    waldur_offering.uuid,
                    object_type,
                )
                continue

            stomp_connections.append((connection, event_subscription, waldur_offering))
        except (errors.UnexpectedStatus, httpx.TimeoutException) as e:
            logger.exception(
                "Unable to register event subscription for offering %s object type %s: %s",
                waldur_offering.name,
                object_type,
                e,
            )

    return stomp_connections


def setup_mqtt_offering_subscriptions(
    waldur_offering: common_structures.Offering, waldur_user_agent: str, global_proxy: str = ""
) -> list[MqttConsumer]:
    """Set up MQTT subscriptions for the specified offering."""
    object_type_to_handler = {}

    if waldur_offering.order_processing_backend:
        object_type_to_handler[ObservableObjectTypeEnum.ORDER] = handlers.on_order_message_mqtt
    else:
        logger.info(
            "Order processing is disabled for offering %s, skipping start of MQTT consumers",
            waldur_offering.name,
        )

    if waldur_offering.membership_sync_backend:
        object_type_to_handler.update(
            {
                ObservableObjectTypeEnum.USER_ROLE: handlers.on_user_role_message_mqtt,
                ObservableObjectTypeEnum.RESOURCE: handlers.on_resource_message_mqtt,
                ObservableObjectTypeEnum.SERVICE_ACCOUNT: handlers.on_account_message_mqtt,
                ObservableObjectTypeEnum.COURSE_ACCOUNT: handlers.on_account_message_mqtt,
            }
        )
    else:
        logger.info(
            "Membership sync is disabled for offering %s, skipping start of MQTT consumers",
            waldur_offering.name,
        )

    waldur_rest_client = get_client(
        waldur_offering.api_url,
        waldur_offering.api_token,
        waldur_user_agent,
        verify_ssl=waldur_offering.verify_ssl,
        proxy=global_proxy,
    )
    event_subscriptions: list[MqttConsumer] = []
    agent_identity_manager = agent_identity_management.AgentIdentityManager(
        waldur_offering, waldur_rest_client
    )
    identity_name = f"agent-{waldur_offering.uuid}"

    try:
        agent_identity = agent_identity_manager.register_identity(identity_name)
    except (errors.UnexpectedStatus, httpx.TimeoutException) as e:
        logger.exception(
            "Failed to register identity for the offering %s: %s", waldur_offering.name, e
        )
        return event_subscriptions

    for object_type, on_message_handler in object_type_to_handler.items():
        try:
            event_subscription = agent_identity_manager.register_event_subscription(
                agent_identity, object_type
            )
            event_subscription_manager = EventSubscriptionManager(
                waldur_offering, on_connect, on_message_handler, waldur_user_agent, object_type
            )
            consumer = event_subscription_manager.start_mqtt_consumer(event_subscription)
            if consumer is None:
                logger.error(
                    "Failed to start mqtt consumer for the offering %s (%s), object type %s",
                    waldur_offering.name,
                    waldur_offering.uuid,
                    object_type,
                )
                event_subscription_manager.delete_event_subscription(event_subscription)
                continue

            event_subscriptions.append((consumer, event_subscription, waldur_offering))
        except (errors.UnexpectedStatus, httpx.TimeoutException) as e:
            logger.exception(
                "Unable to register event subscription for offering %s object type %s: %s",
                waldur_offering.name,
                object_type,
                e,
            )

    return event_subscriptions


def start_stomp_consumers(
    waldur_offerings: list[common_structures.Offering],
    waldur_user_agent: str,
    global_proxy: str = "",
) -> StompConsumersMap:
    """Start multiple STOMP consumers."""
    stomp_consumers_map: StompConsumersMap = {}
    for waldur_offering in waldur_offerings:
        if not waldur_offering.stomp_enabled:
            logger.info("STOMP feature is disabled for the offering")
            continue

        logger.info("Starting STOMP consumers for offering %s", waldur_offering.name)
        stomp_connections = setup_stomp_offering_subscriptions(
            waldur_offering, waldur_user_agent, global_proxy
        )
        if stomp_connections:
            stomp_consumers_map[(waldur_offering.name, waldur_offering.uuid)] = stomp_connections

    return stomp_consumers_map


def stop_stomp_consumers(
    stomp_consumers_map: StompConsumersMap,
) -> None:
    """Stop STOMP consumers."""
    for (offering_name, offering_uuid), consumers in stomp_consumers_map.items():
        logger.info("Stopping STOMP connections for %s (%s)", offering_name, offering_uuid)
        for (
            connection,
            event_subscription,
            offering,
        ) in consumers:
            try:
                event_subscription_manager = EventSubscriptionManager(
                    offering,
                )
                logger.info(
                    "Stopping STOMP connection for %s (%s), observable object type: %s",
                    offering_name,
                    offering_uuid,
                    event_subscription.observable_objects[0]["object_type"]
                    if event_subscription.observable_objects
                    else "N/A",
                )
                event_subscription_manager.stop_stomp_connection(connection)
            except Exception as exc:
                logger.exception("Unable to stop the connection, reason: %s", exc)


def start_mqtt_consumers(
    waldur_offerings: list[common_structures.Offering],
    waldur_user_agent: str,
    global_proxy: str = "",
) -> MqttConsumersMap:
    """Start multiple MQTT consumers."""
    mqtt_consumers_map: MqttConsumersMap = {}
    for waldur_offering in waldur_offerings:
        if not waldur_offering.mqtt_enabled:
            logger.info("MQTT feature is disabled for the offering")
            continue

        logger.info("Setting up MQTT consumers for offering %s", waldur_offering.name)
        event_subscriptions = setup_mqtt_offering_subscriptions(
            waldur_offering, waldur_user_agent, global_proxy
        )
        if event_subscriptions:
            mqtt_consumers_map[(waldur_offering.name, waldur_offering.uuid)] = event_subscriptions

    return mqtt_consumers_map


def stop_mqtt_consumers(
    mqtt_consumers_map: MqttConsumersMap,
) -> None:
    """Stop mqtt consumers and delete event subscriptions."""
    for (offering_name, offering_uuid), subscriptions in mqtt_consumers_map.items():
        logger.info("Stopping subscriptions for %s (%s)", offering_name, offering_uuid)
        for (
            mqttc,
            event_subscription,
            offering,
        ) in subscriptions:
            try:
                event_subscription_manager = EventSubscriptionManager(
                    offering,
                )
                logger.info(
                    "Stopping MQTT consumer for %s (%s), observable object type: %s",
                    offering_name,
                    offering_uuid,
                    event_subscription["observable_objects"][0]["object_type"],
                )
                event_subscription_manager.stop_mqtt_consumer(mqttc)
                logger.info("Deleting event subscription for %s (%s)", offering_name, offering_uuid)
                event_subscription_manager.delete_event_subscription(event_subscription)
            except Exception as exc:
                logger.exception("Unable to stop the consumer, reason: %s", exc)


@contextmanager
def signal_handling(
    mqtt_consumers_map: MqttConsumersMap,
    stomp_consumers_map: StompConsumersMap,
) -> Generator[None, None, None]:
    """Context manager for handling signals gracefully."""

    def signal_handler(signum: int, _: types.FrameType | None) -> None:
        signal_name = signal.Signals(signum).name
        logger.info("Received %s signal. Shutting down gracefully...", signal_name)
        stop_mqtt_consumers(mqtt_consumers_map)
        stop_stomp_consumers(stomp_consumers_map)
        sys.exit(0)

    # Register signal handlers
    signals = (
        signal.SIGTERM,
        signal.SIGINT,
        signal.SIGTSTP,
        signal.SIGQUIT,
    )
    original_handlers = {}

    try:
        logger.info("Setting up signal handlers")
        # Save original handlers and set new ones
        for sig in signals:
            original_handlers[sig] = signal.signal(sig, signal_handler)
        logger.info("Signal handlers set")
        yield
    finally:
        # Restore original handlers
        for sig, handler in original_handlers.items():
            signal.signal(sig, handler)


def run_initial_offering_processing(
    waldur_offerings: list[common_structures.Offering], user_agent: str = ""
) -> None:
    """Runs processing of offerings with MQTT feature enabled."""
    logger.info("Processing offerings with MQTT/STOMP feature enabled")
    for offering in waldur_offerings:
        try:
            if offering.mqtt_enabled or offering.stomp_enabled:
                process_offering(offering, user_agent)
        except Exception as e:
            logger.exception("Error occurred during initial offering process: %s", e)


def process_offering(offering: common_structures.Offering, user_agent: str = "") -> None:
    """Processes the specified offering."""
    logger.info("Processing offering %s (%s)", offering.name, offering.uuid)

    waldur_rest_client = get_client(
        offering.api_url, offering.api_token, user_agent, verify_ssl=offering.verify_ssl
    )
    agent_identity_manager = agent_identity_management.AgentIdentityManager(
        offering, waldur_rest_client
    )
    agent_identity = agent_identity_manager.register_identity(f"agent-{offering.uuid}")
    agent_service = agent_identity_manager.register_service(
        agent_identity,
        "initial-offering-process",
        common_structures.AgentMode.EVENT_PROCESS.value,
    )

    if offering.order_processing_backend:
        order_processor = common_processors.OfferingOrderProcessor(offering, waldur_rest_client)
        order_processor.register(agent_service)
        logger.info("Running offering order process")
        order_processor.process_offering()
    else:
        logger.info("Order processing is disabled for this offering, skipping it")

    if offering.membership_sync_backend:
        membership_processor = common_processors.OfferingMembershipProcessor(
            offering, waldur_rest_client
        )
        membership_processor.register(agent_service)
        logger.info("Running offering membership process")
        membership_processor.process_offering()
    else:
        logger.info("Membership sync is disabled for this offering, skipping it")


def send_agent_health_checks(offerings: list[common_structures.Offering], user_agent: str) -> None:
    """Sends agent health checks for the specified offerings."""
    for offering in offerings:
        try:
            waldur_rest_client = get_client(
                offering.api_url, offering.api_token, user_agent, verify_ssl=offering.verify_ssl
            )
            processor = common_processors.OfferingOrderProcessor(offering, waldur_rest_client)
            marketplace_orders_list.sync(
                client=processor.waldur_rest_client, offering_uuid=offering.uuid
            )
        except Exception as e:
            logger.error(
                "Failed to send agent health check for the offering %s: %s", offering.name, e
            )



"""Handlers for different events and protocols."""

import json

import paho.mqtt.client as mqtt
import stomp
import stomp.utils
from stomp.constants import HDR_DESTINATION
from waldur_api_client import AuthenticatedClient
from waldur_api_client.models import OrderState
from waldur_api_client.models.agent_service import AgentService

from waldur_site_agent.backend import logger
from waldur_site_agent.common import agent_identity_management, structures
from waldur_site_agent.common import processors as common_processors
from waldur_site_agent.common import utils as common_utils
from waldur_site_agent.event_processing.structures import (
    AccountMessage,
    BackendResourceRequestMessage,
    OrderMessage,
    ResourceMessage,
    UserData,
    UserRoleMessage,
)


def register_event_process_service(
    offering: structures.Offering, waldur_rest_client: AuthenticatedClient
) -> AgentService:
    """A shortcut for initialization of the event_process service.

    Args:
        offering (structures.Offering): Waldur offering
        waldur_rest_client (AuthenticatedClient): Waldur API client

    Returns:
        AgentService: Registered agent service
    """
    agent_identity_manager = agent_identity_management.AgentIdentityManager(
        offering, waldur_rest_client
    )
    agent_identity_name = f"agent-{offering.uuid}"
    agent_identity = agent_identity_manager.get_identity(agent_identity_name)
    return agent_identity_manager.register_service(
        agent_identity,
        structures.AgentMode.EVENT_PROCESS.value,
        structures.AgentMode.EVENT_PROCESS.value,
    )


def on_order_message_mqtt(client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage) -> None:
    """Order-processing handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: OrderMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]

    order_uuid = message["order_uuid"]
    order_state = message.get("order_state", "")

    # Skip done and erred orders to avoid duplicate processing
    if order_state in [OrderState.DONE, OrderState.ERRED]:
        logger.info("Skipping order %s with finished state %s", order_uuid, order_state)
        return

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )
        agent_service = register_event_process_service(offering, waldur_rest_client)

        # Create backend instance for dependency injection
        resource_backend, resource_backend_version = common_utils.get_backend_for_offering(
            offering, "order_processing_backend"
        )

        processor = common_processors.OfferingOrderProcessor(
            offering,
            waldur_rest_client,
            resource_backend=resource_backend,
            resource_backend_version=resource_backend_version,
        )
        processor.register(agent_service)

        order = processor.get_order_info(order_uuid)
        if order is None:
            logger.error("Failed to process order %s", order_uuid)
            return
        processor.process_order_with_retries(order)
    except Exception as e:
        logger.error("Failed to process order %s: %s", order_uuid, e)


def on_user_role_message_mqtt(
    client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage
) -> None:
    """Membership sync handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: UserRoleMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]
    user_uuid = message.get("user_uuid")
    project_name = message["project_name"]
    project_uuid = message["project_uuid"]

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )
        agent_service = register_event_process_service(offering, waldur_rest_client)

        # Create backend instance for dependency injection
        resource_backend, resource_backend_version = common_utils.get_backend_for_offering(
            offering, "membership_sync_backend"
        )

        processor = common_processors.OfferingMembershipProcessor(
            offering,
            waldur_rest_client,
            resource_backend=resource_backend,
            resource_backend_version=resource_backend_version,
        )
        processor.register(agent_service)
        if user_uuid:
            user_username = message["user_username"]
            role_granted = message["granted"]
            if role_granted is not None:
                logger.info(
                    "Processing %s (%s) user role changed event in project %s, granted: %s",
                    user_username,
                    user_uuid,
                    project_name,
                    role_granted,
                )
                processor.process_user_role_changed(user_uuid, project_uuid, role_granted)
        else:
            logger.info(
                "Processing full project all users sync event for project %s",
                project_name,
            )
            processor.process_project_user_sync(project_uuid)
    except Exception as e:
        if user_uuid:
            logger.error(
                "Failed to process user %s (%s) role change in project %s (%s) (granted: %s): %s",
                user_username,
                user_uuid,
                project_name,
                project_uuid,
                role_granted,
                e,
            )
        else:
            logger.error(
                "Failed to process full project all users sync event for project %s: %s",
                project_uuid,
                e,
            )


def on_resource_message_mqtt(
    client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage
) -> None:
    """Resource update handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: ResourceMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]
    resource_uuid = message["resource_uuid"]

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )

        agent_service = register_event_process_service(offering, waldur_rest_client)

        processor = common_processors.OfferingMembershipProcessor(offering, waldur_rest_client)
        processor.register(agent_service)
        processor.process_resource_by_uuid(resource_uuid)
    except Exception as e:
        logger.error("Failed to process resource %s: %s", resource_uuid, e)


def process_account_message(
    message: AccountMessage,
    offering: structures.Offering,
    account_type: structures.AccountType,
    user_agent: str = "",
) -> None:
    """Process generic account message."""
    account_username = message["account_username"]
    service_account_uuid = message["account_uuid"]
    project_uuid = message["project_uuid"]
    action = message.get("action", "create")
    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )

        agent_service = register_event_process_service(offering, waldur_rest_client)

        processor = common_processors.OfferingMembershipProcessor(offering, waldur_rest_client)
        processor.register(agent_service)
        if action == "create":
            processor.process_account_creation(account_username, account_type)
        elif action == "delete":
            processor.process_account_removal(account_username, project_uuid)
        else:
            logger.error("Unknown action %s for course account %s", action, account_username)
    except Exception as e:
        logger.error(
            "Failed to process %s of course account %s (%s): %s",
            action,
            account_username,
            service_account_uuid,
            e,
        )


def on_account_message_mqtt(client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage) -> None:
    """Generic account handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: AccountMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]
    account_type_raw = msg.topic.split("/")[-1]
    account_type = structures.AccountType.SERVICE_ACCOUNT
    if account_type_raw == structures.AccountType.COURSE_ACCOUNT.value:
        account_type = structures.AccountType.COURSE_ACCOUNT
    process_account_message(message, offering, account_type, user_agent)


def on_order_message_stomp(
    frame: stomp.utils.Frame, offering: structures.Offering, user_agent: str
) -> None:
    """Order-processing handler for STOMP message event."""
    message: OrderMessage = json.loads(frame.body)
    logger.info("Processing message: %s", message)
    order_uuid = message["order_uuid"]
    order_state = message.get("order_state", "")

    # Skip done and erred orders to avoid duplicate processing
    if order_state in [OrderState.DONE, OrderState.ERRED]:
        logger.info("Skipping order %s with finished state %s", order_uuid, order_state)
        return

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )
        agent_service = register_event_process_service(offering, waldur_rest_client)

        # Create backend instance for dependency injection
        resource_backend, resource_backend_version = common_utils.get_backend_for_offering(
            offering, "order_processing_backend"
        )

        processor = common_processors.OfferingOrderProcessor(
            offering,
            waldur_rest_client,
            resource_backend=resource_backend,
            resource_backend_version=resource_backend_version,
        )
        processor.register(agent_service)

        order = processor.get_order_info(order_uuid)
        if order is None:
            logger.error("Failed to process order %s", order_uuid)
            return
        processor.process_order_with_retries(order)
    except Exception as e:
        logger.error("Failed to process order %s: %s", order_uuid, e)


def on_user_role_message_stomp(
    frame: stomp.utils.Frame, offering: structures.Offering, user_agent: str
) -> None:
    """Membership sync handler for STOMP message event."""
    message: UserRoleMessage = json.loads(frame.body)
    logger.info("Received message: %s on topic %s", message, frame.headers.get("destination"))
    user_uuid = message.get("user_uuid")
    project_name = message["project_name"]
    project_uuid = message["project_uuid"]

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )
        agent_service = register_event_process_service(offering, waldur_rest_client)

        # Create backend instance for dependency injection
        resource_backend, resource_backend_version = common_utils.get_backend_for_offering(
            offering, "membership_sync_backend"
        )

        processor = common_processors.OfferingMembershipProcessor(
            offering,
            waldur_rest_client,
            resource_backend=resource_backend,
            resource_backend_version=resource_backend_version,
        )
        processor.register(agent_service)
        if user_uuid:
            user_username = message["user_username"]
            role_granted = message["granted"]
            if role_granted is None:
                logger.error("Missing required field 'granted' for user role change")
                return
            logger.info(
                "Processing %s (%s) user role changed event in project %s, granted: %s",
                user_username,
                user_uuid,
                project_name,
                role_granted,
            )
            processor.process_user_role_changed(user_uuid, project_uuid, role_granted)
        else:
            logger.info("Processing full project all users sync event for project %s", project_name)
            processor.process_project_user_sync(project_uuid)
    except Exception as e:
        if user_uuid:
            logger.error(
                "Failed to process user %s (%s) role change in project %s (%s) (granted: %s): %s",
                user_username,
                user_uuid,
                project_name,
                project_uuid,
                role_granted,
                e,
            )
        else:
            logger.error(
                "Failed to process full project all users sync event for project %s: %s",
                project_uuid,
                e,
            )


def on_resource_message_stomp(
    frame: stomp.utils.Frame, offering: structures.Offering, user_agent: str
) -> None:
    """Resource update handler for STOMP message event."""
    message: ResourceMessage = json.loads(frame.body)
    resource_uuid = message["resource_uuid"]

    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )

        agent_service = register_event_process_service(offering, waldur_rest_client)
        processor = common_processors.OfferingMembershipProcessor(offering, waldur_rest_client)
        processor.register(agent_service)

        processor.process_resource_by_uuid(resource_uuid)
    except Exception as e:
        logger.error("Failed to process resource %s: %s", resource_uuid, e)


def on_importable_resources_message_stomp(
    frame: stomp.utils.Frame, offering: structures.Offering, user_agent: str
) -> None:
    """Handler for importable resource list request for STOMP message event."""
    message: BackendResourceRequestMessage = json.loads(frame.body)
    request_uuid = message["backend_resource_request_uuid"]
    try:
        waldur_rest_client = common_utils.get_client(
            offering.api_url, offering.api_token, user_agent, offering.verify_ssl
        )

        agent_service = register_event_process_service(offering, waldur_rest_client)

        # Create backend instance for dependency injection
        resource_backend, resource_backend_version = common_utils.get_backend_for_offering(
            offering, "order_processing_backend"
        )

        processor = common_processors.OfferingImportableResourcesProcessor(
            offering,
            waldur_rest_client,
            resource_backend=resource_backend,
            resource_backend_version=resource_backend_version,
        )
        processor.register(agent_service)

        processor.process_request(request_uuid)
    except Exception as e:
        logger.error("Failed to process importable resource list request %s: %s", request_uuid, e)


def on_account_message_stomp(
    frame: stomp.utils.Frame, offering: structures.Offering, user_agent: str
) -> None:
    """Service account handler for STOMP."""
    message: AccountMessage = json.loads(frame.body)
    queue: str = frame.headers[HDR_DESTINATION]
    queue_parts = queue.split("_")
    account_type_raw = f"{queue_parts[-2]}_{queue_parts[-1]}"
    account_type = structures.AccountType.SERVICE_ACCOUNT
    if account_type_raw == structures.AccountType.COURSE_ACCOUNT.value:
        account_type = structures.AccountType.COURSE_ACCOUNT
    process_account_message(message, offering, account_type, user_agent)
